import json
from contextlib import nullcontext
from typing import Optional
from typing import Union

import click
from click import BadOptionUsage
from click import File
from rich.panel import Panel

from _incydr_cli import console
from _incydr_cli import get_user_project_path
from _incydr_cli import logging_options
from _incydr_cli import render
from _incydr_cli.cmds.options.event_filter_options import advanced_query_option
from _incydr_cli.cmds.options.event_filter_options import event_filter_options
from _incydr_cli.cmds.options.event_filter_options import saved_search_option
from _incydr_cli.cmds.options.output_options import columns_option
from _incydr_cli.cmds.options.output_options import output_options
from _incydr_cli.cmds.options.output_options import single_format_option
from _incydr_cli.cmds.options.output_options import SingleFormat
from _incydr_cli.cmds.options.output_options import table_format_option
from _incydr_cli.cmds.options.output_options import TableFormat
from _incydr_cli.cmds.options.utils import checkpoint_option
from _incydr_cli.cmds.utils import warn_interrupt
from _incydr_cli.core import IncydrCommand
from _incydr_cli.core import IncydrGroup
from _incydr_cli.cursor import CursorStore
from _incydr_cli.logger import get_server_logger
from _incydr_sdk.core.client import Client
from _incydr_sdk.enums.file_events import RiskIndicators
from _incydr_sdk.enums.file_events import RiskSeverity
from _incydr_sdk.file_events.models.event import FileEventV2
from _incydr_sdk.file_events.models.response import SavedSearch
from _incydr_sdk.queries.file_events import EventQuery
from _incydr_sdk.utils import model_as_card


@click.group(cls=IncydrGroup)
@logging_options
def file_events():
    """View and manage file events."""


@file_events.command(cls=IncydrCommand)
@checkpoint_option
@table_format_option
@columns_option
@output_options
@advanced_query_option
@saved_search_option
@event_filter_options
@logging_options
def search(
    format_: TableFormat,
    columns: Optional[str],
    output: Optional[str],
    certs: Optional[str],
    ignore_cert_validation: Optional[bool],
    advanced_query: Optional[Union[str, File]],
    saved_search: Optional[str],
    start: Optional[str],
    end: Optional[str],
    event_action: Optional[str],
    username: Optional[str],
    md5: Optional[str],
    sha256: Optional[str],
    source_category: Optional[str],
    destination_category: Optional[str],
    file_name: Optional[str],
    file_directory: Optional[str],
    file_category: Optional[str],
    risk_indicator: Optional[RiskIndicators],
    risk_severity: Optional[RiskSeverity],
    risk_score: Optional[int],
    checkpoint_name: Optional[str],
):
    """
    Search file events. Various options are provided to filter query results.

    Use the `--saved-search` or the `--advanced-query` option if the available filters don't satisfy your requirements.

    Defaults to returning events with a risk score >= 1.  Add the `--risk-score 0` filter to return all events,
    including those with no risk associated with them.

    Results will be output to the console by default, use the `--output` option to send data to a server.

    Checkpointing is available through the `--checkpoint <checkpoint-name>` option and will only return new results
    on subsequent queries with that same checkpoint.  Checkpointing stores the original query it was run with, so
    additional filters on subsequent runs will be ignored.
    """
    if output:
        format_ = TableFormat.json_lines

    client = Client()

    if saved_search:
        saved_search = client.file_events.v2.get_saved_search(saved_search)
        query = EventQuery.from_saved_search(saved_search)
    elif advanced_query:
        if not isinstance(advanced_query, str):
            advanced_query = advanced_query.read()
        query = EventQuery.parse_raw(advanced_query)
    else:
        if not start:
            raise BadOptionUsage(
                "start",
                "--start option required if not using --saved-search or --advanced-query options.",
            )
        query = _create_query(
            start=start,
            end=end,
            event_action=event_action,
            username=username,
            md5=md5,
            sha256=sha256,
            source_category=source_category,
            destination_category=destination_category,
            file_name=file_name,
            file_directory=file_directory,
            file_category=file_category,
            risk_indicator=risk_indicator,
            risk_severity=risk_severity,
            risk_score=risk_score,
        )

    query.page_size = 10000

    if checkpoint_name:
        cursor = _get_cursor_store(client.settings.api_client_id)
        checkpoint = cursor.get(checkpoint_name)
        if checkpoint:  # if stored checkpoint, overwrite query
            query = EventQuery.parse_raw(checkpoint)

        def checkpoint_func(event_):
            # Stored checkpoints are json strings of the original query with the `pageToken` key
            # updated to contain the ID of the last returned event
            q = query.copy()
            q.page_token = event_["event"]["id"]
            cursor.replace(checkpoint_name, json.dumps(q.dict()))

    else:
        checkpoint_func = None

    # skip pydantic modeling when output will just be json
    if format_ in (TableFormat.json_pretty, TableFormat.json_lines):

        def yield_all_events(q: EventQuery):
            while q.page_token is not None:
                response = client.session.post("/v2/file-events", json=q.dict())
                response_dict = response.json()
                q.page_token = response_dict.get("nextPgToken")
                page = response_dict.get("fileEvents")
                for event_ in page:
                    yield event_
                    if checkpoint_func:
                        checkpoint_func(event_)

    else:

        def yield_all_events(q: EventQuery):
            while q.page_token is not None:
                page = client.file_events.v2.search(q).file_events
                for event_ in page:
                    yield event_
                    if checkpoint_func:
                        checkpoint_func(event_.dict())

    events = yield_all_events(query)

    with warn_interrupt() if checkpoint_name else nullcontext():
        if output:
            logger = get_server_logger(output, certs, ignore_cert_validation)
            for event in events:
                logger.info(json.dumps(event))
            return

        if format_ == TableFormat.csv:
            render.csv(FileEventV2, events, columns=columns, flat=True)
        elif format_ == TableFormat.table:
            render.table(FileEventV2, events, columns=columns, flat=False)
        else:
            printed = False
            for event in events:
                printed = True
                if format_ == TableFormat.json_pretty:
                    console.print_json(data=event)
                else:
                    click.echo(json.dumps(event))
            if not printed:
                console.print("No results found.")


@file_events.command()
@click.argument("checkpoint-name")
def clear_checkpoint(checkpoint_name: str):
    """Remove the saved file events checkpoint from searches made with `--checkpoint` mode."""
    client = Client()
    cursor = _get_cursor_store(client.settings.api_client_id)
    cursor.delete(checkpoint_name)


@file_events.command(cls=IncydrCommand)
@single_format_option
@click.argument("search-id")
@logging_options
def show_saved_search(
    search_id: str,
    format_: SingleFormat,
):
    """
    Show details for a single saved search.
    """
    client = Client()
    saved_search = client.file_events.v2.get_saved_search(search_id)
    if format_ == SingleFormat.rich:
        console.print(Panel.fit(model_as_card(saved_search)))
    elif format_ == SingleFormat.json_pretty:
        console.print_json(saved_search.json())
    else:
        click.echo(saved_search.json())


@file_events.command(cls=IncydrCommand)
@table_format_option
@columns_option
@logging_options
def list_saved_searches(
    format_: TableFormat,
    columns: Optional[str],
):
    """
    List saved searches.
    """
    client = Client()
    searches = client.file_events.v2.list_saved_searches()
    if format_ == TableFormat.table:
        render.table(SavedSearch, searches, columns=columns, flat=False)
    elif format_ == TableFormat.csv:
        render.csv(SavedSearch, searches, columns=columns, flat=True)
    elif format_ == TableFormat.json_pretty:
        for s in searches:
            console.print_json(s.json())
    else:
        for s in searches:
            click.echo(s.json())


field_option_map = {
    "event_action": "event.action",
    "username": "user.email",
    "md5": "file.hash.md5",
    "sha256": "file.hash.sha256",
    "source_category": "source.category",
    "destination_category": "destination.category",
    "file_name": "file.name",
    "file_directory": "file.directory",
    "risk_indicator": "risk.indicators.name",
    "file_category": "file.category",
    "risk_severity": "risk.severity",
    "risk_score": "risk.score",
}


def _create_query(**kwargs):
    query = EventQuery(start_date=kwargs["start"], end_date=kwargs["end"])
    for k, v in kwargs.items():
        if v:
            if k in ["start", "end"]:
                continue
            elif k == "risk_score":
                query = query.greater_than(field_option_map[k], v)
            else:
                query = query.equals(field_option_map[k], v)
    return query


def _get_cursor_store(api_key):
    """
    Get cursor store for file event search checkpoints.
    """
    dir_path = get_user_project_path(
        "checkpoints",
        api_key,
        "file_event_checkpoints",
    )
    return CursorStore(dir_path, "file_events")


# Allows us to import individual command groups
if __name__ == "__main__":
    file_events()



from _incydr_sdk.enums import _Enum


class WatchlistType(_Enum):
    """
    Available watchlist types.
    """

    WATCHLIST_TYPE_UNSPECIFIED = "WATCHLIST_TYPE_UNSPECIFIED"
    CONTRACT_EMPLOYEE = "CONTRACT_EMPLOYEE"
    DEPARTING_EMPLOYEE = "DEPARTING_EMPLOYEE"
    ELEVATED_ACCESS_PRIVILEGES = "ELEVATED_ACCESS_PRIVILEGES"
    FLIGHT_RISK = "FLIGHT_RISK"
    HIGH_IMPACT_EMPLOYEE = "HIGH_IMPACT_EMPLOYEE"
    NEW_EMPLOYEE = "NEW_EMPLOYEE"
    PERFORMANCE_CONCERNS = "PERFORMANCE_CONCERNS"
    POOR_SECURITY_PRACTICES = "POOR_SECURITY_PRACTICES"
    SUSPICIOUS_SYSTEM_ACTIVITY = "SUSPICIOUS_SYSTEM_ACTIVITY"
    CUSTOM = "CUSTOM"



import click
from rich.panel import Panel

from _incydr_cli import console
from _incydr_cli import logging_options
from _incydr_cli import render
from _incydr_cli.cmds.options.output_options import columns_option
from _incydr_cli.cmds.options.output_options import single_format_option
from _incydr_cli.cmds.options.output_options import SingleFormat
from _incydr_cli.cmds.options.output_options import table_format_option
from _incydr_cli.cmds.options.output_options import TableFormat
from _incydr_cli.cmds.utils import deprecation_warning
from _incydr_cli.core import IncydrCommand
from _incydr_cli.core import IncydrGroup
from _incydr_sdk.core.client import Client
from _incydr_sdk.devices.models import Device
from _incydr_sdk.utils import model_as_card


# Deprecated 2025-03
DEPRECATION_TEXT = "DeprecationWarning: Devices commands are deprecated. Use the 'incydr agents' command group instead."


@click.group(cls=IncydrGroup)
@logging_options
def devices():
    """DEPRECATED. Use the agents command group instead. View devices."""
    deprecation_warning(DEPRECATION_TEXT)


@devices.command("list", cls=IncydrCommand)
@click.option(
    "--active/--inactive",
    default=None,
    help="Filter by active or inactive devices. Defaults to returning both when when neither option is passed.",
)
@click.option(
    "--blocked/--unblocked",
    default=None,
    help="Filter by blocked or unblocked devices. Defaults to returning both when when neither option is passed.",
)
@table_format_option
@columns_option
@logging_options
def list_(
    active: bool = None,
    blocked: bool = None,
    format_: TableFormat = None,
    columns: str = None,
):
    """
    List devices.
    """
    client = Client()
    devices = client.devices.v1.iter_all(active, blocked)

    if format_ == TableFormat.table:
        columns = columns or [
            "device_id",
            "name",
            "os_hostname",
            "status",
            "active",
            "blocked",
            "alert_state",
            "org_guid",
            "login_date",
        ]
        render.table(Device, devices, columns=columns, flat=False)
    elif format_ == TableFormat.csv:
        render.csv(Device, devices, columns=columns, flat=True)
    elif format_ == TableFormat.json_pretty:
        for device in devices:
            console.print_json(device.json())
    else:
        for device in devices:
            click.echo(device.json())


@devices.command(cls=IncydrCommand)
@click.argument("device_id")
@single_format_option
@logging_options
def show(
    device_id: str,
    format_: SingleFormat,
):
    """
    Show details for a single device.
    """
    client = Client()
    device = client.devices.v1.get_device(device_id)

    if format_ == SingleFormat.rich and client.settings.use_rich:
        console.print(
            Panel.fit(model_as_card(device), title=f"Device {device.device_id}")
        )
    elif format_ == SingleFormat.json_pretty:
        console.print_json(device.json())
    else:
        click.echo(device.json())



from functools import lru_cache
from pathlib import Path
from typing import Optional

import click
from boltons.iterutils import bucketize
from pydantic import Field
from requests import HTTPError
from rich.panel import Panel
from rich.progress import track

from _incydr_cli import console
from _incydr_cli import logging_options
from _incydr_cli import render
from _incydr_cli.cmds.models import UserCSV
from _incydr_cli.cmds.models import UserJSON
from _incydr_cli.cmds.options.output_options import columns_option
from _incydr_cli.cmds.options.output_options import input_format_option
from _incydr_cli.cmds.options.output_options import single_format_option
from _incydr_cli.cmds.options.output_options import SingleFormat
from _incydr_cli.cmds.options.output_options import table_format_option
from _incydr_cli.cmds.options.output_options import TableFormat
from _incydr_cli.cmds.options.utils import user_lookup_callback
from _incydr_cli.cmds.utils import deprecation_warning
from _incydr_cli.cmds.utils import user_lookup
from _incydr_cli.core import IncydrCommand
from _incydr_cli.core import IncydrGroup
from _incydr_cli.file_readers import AutoDecodedFile
from _incydr_sdk.agents.models import Agent
from _incydr_sdk.core.client import Client
from _incydr_sdk.devices.models import Device
from _incydr_sdk.users.client import RoleNotFoundError
from _incydr_sdk.users.client import RoleProcessingError
from _incydr_sdk.users.client import UserNotAssignedRoleError
from _incydr_sdk.users.models import Role
from _incydr_sdk.users.models import User
from _incydr_sdk.users.models import UserRole
from _incydr_sdk.utils import model_as_card

user_arg = click.argument("user", callback=user_lookup_callback)


@click.group(cls=IncydrGroup)
@logging_options
def users():
    """View and manage users and user roles."""


@users.command("list", cls=IncydrCommand)
@click.option(
    "--active/--inactive",
    default=None,
    help="Filter by active or inactive users. Defaults to returning both when when neither option is passed.",
)
@click.option(
    "--blocked/--unblocked",
    default=None,
    help="Filter by blocked or unblocked users. Defaults to returning both when when neither option is passed.",
)
@click.option("--username", default=None)
@table_format_option
@columns_option
@logging_options
def list_(
    active: Optional[bool],
    blocked: Optional[bool],
    username: Optional[str],
    format_: TableFormat,
    columns: Optional[str],
):
    """
    List users.
    """
    client = Client()
    users_ = client.users.v1.iter_all(active=active, blocked=blocked, username=username)

    if format_ == TableFormat.csv:
        render.csv(User, users_, columns=columns, flat=True)
    elif format_ == TableFormat.table:
        columns = columns or [
            "user_id",
            "username",
            "first_name",
            "last_name",
            "org_guid",
            "org_name",
            "notes",
            "active",
            "blocked",
        ]
        render.table(User, users_, columns=columns, flat=False)
    else:
        printed = False
        if format_ == TableFormat.json_pretty:
            for item in users_:
                printed = True
                console.print_json(item.json())
        else:
            for item in users_:
                printed = True
                click.echo(item.json())
        if not printed:
            console.print("No results found.")


@users.command(cls=IncydrCommand)
@click.argument("user")  # don't need username callback, handled by client method
@single_format_option
@logging_options
def show(
    user,
    format_: SingleFormat,
):
    """
    Show details for a user.

    Accepts a user ID or a username.  Performs an additional lookup if username is passed.
    """
    client = Client()
    user = client.users.v1.get_user(user)

    if format_ == SingleFormat.rich and client.settings.use_rich:
        console.print(Panel.fit(model_as_card(user), title=f"User {user.username}"))
    elif format_ == SingleFormat.json_pretty:
        console.print_json(user.json())
    else:
        click.echo(user.json())


@users.command(cls=IncydrCommand)
@user_arg
@table_format_option
@columns_option
@logging_options
def list_devices(
    user,
    format_: TableFormat,
    columns: Optional[str],
):
    """
    DEPRECATED - use list-agents instead.
    """
    deprecation_warning("DEPRECATED. Use list-agents instead.")
    client = Client()
    devices = client.users.v1.get_devices(user).devices

    if format_ == TableFormat.csv:
        render.csv(Device, devices, columns=columns, flat=True)
    elif format_ == TableFormat.table:
        columns = columns or [
            "device_id",
            "name",
            "os_hostname",
            "status",
            "active",
            "blocked",
            "alert_state",
            "org_guid",
            "login_date",
        ]
        render.table(Device, devices, columns=columns, flat=False)
    elif format_ == TableFormat.json_pretty:
        for item in devices:
            console.print_json(item.json())
    else:
        for item in devices:
            click.echo(item.json())


@users.command(cls=IncydrCommand)
@user_arg
@table_format_option
@columns_option
@logging_options
def list_agents(
    user,
    format_: TableFormat,
    columns: Optional[str],
):
    """
    List agents associated with a particular user.
    """
    client = Client()
    if "@" in user:
        user = client.users.v1.get_user(user).user_id
    devices = list(client.agents.v1.iter_all(user_id=user))

    if format_ == TableFormat.csv:
        render.csv(Agent, devices, columns=columns, flat=True)
    elif format_ == TableFormat.table:
        columns = columns or [
            "agent_id",
            "name",
            "os_hostname",
            "active",
            "agent_type",
            "agent_health_issue_types",
            "last_connected",
        ]
        render.table(Agent, devices, columns=columns, flat=False)
    elif format_ == TableFormat.json_pretty:
        for item in devices:
            console.print_json(item.json())
    else:
        for item in devices:
            click.echo(item.json())


@users.command("list-roles", cls=IncydrCommand)
@user_arg
@table_format_option
@columns_option
@logging_options
def list_user_roles(
    user: str,
    format_: TableFormat,
    columns: Optional[str],
):
    """
    List roles associated with a particular user.
    """
    client = Client()
    roles = client.users.v1.list_user_roles(user)

    if format_ == TableFormat.csv:
        render.csv(UserRole, roles, columns=columns, flat=True)
    elif format_ == TableFormat.table:
        render.table(UserRole, roles, columns=columns, flat=False)
    elif format_ == TableFormat.json_pretty:
        for item in roles:
            console.print_json(item.json())
    else:
        for item in roles:
            click.echo(item.json())


@users.command(cls=IncydrCommand)
@user_arg
@click.argument("roles")
@click.option("--add", "update_method", flag_value="add", default=None)
@click.option("--remove", "update_method", flag_value="remove", default=None)
@logging_options
def update_roles(
    user,
    roles: str,
    update_method: str,
):
    """
    Update roles associated with a particular user.

    Usage:
        users update-roles USER ROLES

    USER is the user ID or username of the user whose roles will be updated.
    Performs an additional lookup if username is passed.

    ROLES is a comma-delimited list of role IDs and/or role names to replace that user's roles.

    Use the `--remove` flag to remove the specified role(s) from a user's existing roles.

    Alternatively, use the `--add` flag to assign additional roles to a user's existing roles.
    """
    client = Client()

    if update_method == "add":
        update_func = client.users.v1.add_roles
    elif update_method == "remove":
        update_func = client.users.v1.remove_roles
    else:
        update_func = client.users.v1.update_roles

    try:
        update_func(user, [r.strip() for r in roles.split(",")])
    except RoleNotFoundError as e:
        raise e
    except UserNotAssignedRoleError as e:
        raise e
    console.print(f"Roles successfully updated for user '{user}'")


@users.command(cls=IncydrCommand)
@user_arg
@logging_options
def activate(
    user,
):
    """
    Activate a user.
    """
    client = Client()
    client.users.v1.activate(user)
    console.print(f"User '{user}' successfully activated.")


@users.command(cls=IncydrCommand)
@user_arg
@logging_options
def deactivate(
    user,
):
    """
    Deactivate a user.
    """
    client = Client()
    client.users.v1.deactivate(user)
    console.print(f"User '{user}' successfully deactivated.")


@users.command(cls=IncydrCommand)
@user_arg
@click.argument("org_guid")
@logging_options
def move(
    user,
    org_guid,
):
    """
    Move a user to a specified organization.
    """
    client = Client()
    client.users.v1.move(user, org_guid)
    console.print(f"User '{user}' successfully moved to org '{org_guid}'.")


@users.group("roles", cls=IncydrGroup)
def roles_():
    """View available roles."""


@roles_.command("show", cls=IncydrCommand)
@single_format_option
@click.argument("role")
@logging_options
def show_role(
    role,
    format_: SingleFormat,
):
    """
    Show details for a single role, specified by role name or role ID.
    """
    client = Client()
    role = client.users.v1.get_role(role)
    if format_ == SingleFormat.rich and client.settings.use_rich:
        console.print(Panel.fit(model_as_card(role), title=role.role_name))
    elif format_ == SingleFormat.json_pretty:
        console.print_json(role.json())
    else:
        click.echo(role.json())


@roles_.command("list", cls=IncydrCommand)
@table_format_option
@logging_options
def list_roles(
    format_: TableFormat,
):
    """
    List all available roles that can be assigned by the current user.
    """
    client = Client()
    roles = client.users.v1.list_roles()
    if format_ == TableFormat.csv:
        render.csv(Role, roles, flat=True)
    elif format_ == TableFormat.table:
        render.table(Role, roles, flat=False)
    elif format_ == TableFormat.json_pretty:
        for item in roles:
            console.print_json(item.json())
    else:
        for item in roles:
            click.echo(item.json())


@users.command(cls=IncydrCommand)
@click.argument("file", type=AutoDecodedFile())
@input_format_option
@click.option("--add", "update_method", flag_value="add", default=None)
@click.option("--remove", "update_method", flag_value="remove", default=None)
@logging_options
def bulk_update_roles(
    file,
    update_method: str,
    format_: Optional[str],
):
    """
    Bulk update roles associated with multiple users from a file.

    By default, the provided roles will replace the specified user's existing roles.
    Use the `--add` flag or the `--remove` flag to add or remove roles, respectively, from a user's existing roles.

    Takes a single arg `FILE` which specifies the path to the file (use "-" to read from stdin).

    File format can either be CSV or [JSON Lines format](https://jsonlines.org) (Default is CSV).

    Requires the following CSV columns or JSON keys:

    * `user` - User ID or username of the user whose roles will be updated. Performs an additional lookup if username is passed.
    * `role` - Role ID and/or role name (case-sensitive) to assign to the new user.

    Example:

        The following command will add roles to users as specified in a JSON-lines formatted file:

            incydr users bulk-update-roles path/to/file.json --add --format json-lines
    """
    client = Client()

    @lru_cache()
    def resolve_username(user):
        if user is None:
            return
        elif "@" in user:
            return user_lookup(client, user)
        else:  # assume user_id
            return user

    class RoleUpdateCSV(UserCSV):
        role: str = Field(
            csv_aliases=["role", "role_id", "role_name", "roleId", "roleName"]
        )

    class RoleUpdateJSON(UserJSON):
        role: str = Field(alias="roleId")

    if format_ == "csv":
        roles_file = RoleUpdateCSV.parse_csv(file)
    else:
        roles_file = RoleUpdateJSON.parse_json_lines(file)

    if update_method == "add":
        update_func = client.users.v1.add_roles
    elif update_method == "remove":
        update_func = client.users.v1.remove_roles
    else:
        update_func = client.users.v1.update_roles

    # group updates by user
    buckets = bucketize(
        roles_file,
        key=lambda role_update: role_update.user,
        value_transform=lambda role_update: role_update.role,
    )
    for bucket in buckets:
        user = bucket
        roles = buckets[bucket]
        try:
            user = resolve_username(user)
            update_func(user, roles)
            console.print(f"Successfully updated roles for user {user}.'")
        except RoleProcessingError as err:
            console.print(
                f"[red]Error! {str(err)} [/red]",
                highlight=False,
            )


@users.command(cls=IncydrCommand)
@click.argument("file", type=click.File())
@input_format_option
@logging_options
def bulk_activate(file: Path, format_: str):
    """
    Bulk activate users.

    Takes a single arg `FILE` which specifies the path to the file (use "-" to read from stdin).

    File format can either be CSV or [JSON Lines format](https://jsonlines.org) (Default is CSV).

    Requires a single `user` column or field that contains either the user IDs or the usernames of the users to be activated.
    """
    client = Client()

    @lru_cache()
    def resolve_username(user):
        if user is None:
            return
        elif "@" in user:
            return user_lookup(client, user)
        else:  # assume user_id
            return user

    if format_ == "csv":
        models = UserCSV.parse_csv(file)
    else:
        models = UserJSON.parse_json_lines(file)

    try:
        for row in track(
            models,
            description="Activating users...",
            transient=True,
        ):
            client.users.v1.activate(resolve_username(row.user))
    except HTTPError as err:
        console.print(f"[red]Error:[/red] {err.response.text}")


@users.command(cls=IncydrCommand)
@click.argument("file", type=click.File())
@input_format_option
@logging_options
def bulk_deactivate(file: Path, format_: str):
    """
    Bulk deactivate users.

    Takes a single arg `FILE` which specifies the path to the file (use "-" to read from stdin).

    File format can either be CSV or [JSON Lines format](https://jsonlines.org) (Default is CSV).

    Requires a single `user` column or field that contains either the user IDs or the usernames of the users to be deactivated.
    """
    client = Client()

    @lru_cache()
    def resolve_username(user):
        if user is None:
            return
        elif "@" in user:
            return user_lookup(client, user)
        else:  # assume user_id
            return user

    if format_ == "csv":
        models = UserCSV.parse_csv(file)
    else:
        models = UserJSON.parse_json_lines(file)

    try:
        for row in track(
            models,
            description="Deactivating users...",
            transient=True,
        ):
            client.users.v1.deactivate(resolve_username(row.user))
    except HTTPError as err:
        console.print(f"[red]Error:[/red] {err.response.text}")


@users.command(cls=IncydrCommand)
@click.argument("file", type=click.File())
@input_format_option
@logging_options
def bulk_move(file: Path, format_: str):

    """
    Bulk move multiple users to specified organizations.

    Takes a single arg `FILE` which specifies the path to the file (use "-" to read from stdin).

    File format can either be CSV or [JSON Lines format](https://jsonlines.org) (Default is CSV).

    Requires the following columns:

    * `user` - User ID or username of the user who will be moved to the new organization. Performs an additional lookup if username is passed.
    * `org_guid` - GUID for the user's new organization.
    """
    client = Client()

    class UserMoveCSV(UserCSV):
        org_guid: str = Field(csv_aliases=["org_guid", "orgGuid", "org"])

    class UserMoveJSON(UserJSON):
        org_guid: str = Field(alias="orgGuid")

    @lru_cache()
    def resolve_username(user):
        if user is None:
            return
        elif "@" in user:
            return user_lookup(client, user)
        else:  # assume user_id
            return user

    if format_ == "csv":
        models = UserMoveCSV.parse_csv(file)
    else:
        models = UserMoveJSON.parse_json_lines(file)

    try:
        for row in track(
            models,
            description="Moving users...",
            transient=True,
        ):
            client.users.v1.move(resolve_username(row.user), row.org_guid)
    except HTTPError as err:
        console.print(f"[red]Error:[/red] {err.response.text}")



from __future__ import annotations

from datetime import datetime
from typing import List
from typing import Optional
from typing import Union

from pydantic import BaseModel
from pydantic import Extra
from pydantic import Field
from rich.markdown import Markdown

from _incydr_sdk.core.models import Model
from _incydr_sdk.core.models import ResponseModel
from _incydr_sdk.enums import SortDirection
from _incydr_sdk.enums.cases import CaseStatus
from _incydr_sdk.enums.cases import FileAvailability
from _incydr_sdk.enums.cases import SortKeys
from _incydr_sdk.utils import list_as_panel
from _incydr_sdk.utils import model_as_card


class Case(ResponseModel, validate_assignment=True):
    """A model representing an Incydr Case.

    **Fields**:

    * **number**: `int` Unique numeric identifier for the case.
    * **name**: `str` Unique name given to the case.
    * **created_at**: `datetime` Time at which the case was created. json_alias=`createdAt`
    * **updated_at**: `datetime | None` Time at which the case was last updated. json_alias=`updatedAt`
    * **subject**: `str | None` The user UID of the subject being investigated in this case.
    * **subject_username**: `str | None` The username of the subject being investigated in this case. json_alias=`subjectUsername`
    * **status**: `CaseStatus` Indicates the status of the case. OPEN: The case is active and all aspects of the case are editable. CLOSED: The case is resolved. Closed cases cannot be re-opened or modified. Case data for closed cases is retained indefinitely.
    * **assignee**: `str | None` The user ID of the administrator assigned to investigate the case.
    * **assignee_username**: `str | None` The username of the administrator assigned to investigate the case. json_alias=`assigneeUsername`
    * **created_by_user_id**: `str | None` User UID of the user who created the case. json_alias=`createdByUserUid`
    * **created_by_username**: `str | None` Username of the user who created the case. json_alias=`createdByUsername`
    * **last_modified_by_user_id**: `str | None` User UID of the user who last modified the case. json_alias=`lastModifiedByUserUid`
    * **last_modified_by_username**: `str | None` Username of the user who last modified the case. json_alias=`lastModifiedByUsername`
    * **archival_time**: `datetime` Time at which the case will be archived.
    """

    number: Optional[int] = Field(
        None, frozen=True, description="The identifier of the case."
    )
    name: Optional[str]
    created_at: Optional[datetime] = Field(None, frozen=True, alias="createdAt")
    updated_at: Optional[datetime] = Field(None, frozen=True, alias="updatedAt")
    subject: Optional[str] = None
    subject_username: Optional[str] = Field(None, alias="subjectUsername")
    status: Union[CaseStatus, str] = None
    assignee: Optional[str] = None
    assignee_username: Optional[str] = Field(
        None, frozen=True, alias="assigneeUsername"
    )
    created_by_user_id: Optional[str] = Field(
        None, frozen=True, alias="createdByUserUid"
    )
    created_by_username: Optional[str] = Field(
        None, frozen=True, alias="createdByUsername"
    )
    last_modified_by_user_id: Optional[str] = Field(
        None, frozen=True, alias="lastModifiedByUserUid"
    )
    last_modified_by_username: Optional[str] = Field(
        None, frozen=True, alias="lastModifiedByUsername"
    )
    archival_time: Optional[datetime] = Field(None, frozen=True, alias="archivalTime")


class CaseDetail(Case):
    """A model representing the full details of Incydr Case.

    **Fields**:

    **All the same fields as Case model, plus**:

    * **description**: `str | None` Brief description providing context for a case.
    * **findings**: `str | None` Markdown formatted text summarizing the findings for a case.
    """

    description: Optional[str] = None
    findings: Optional[str] = Field(
        None, table=lambda f: f if f is None else Markdown(f)
    )


class CasesPage(ResponseModel):
    """
    A model representing a page of `Case` objects.

    **Fields**:

    * **cases**: `List[Case]` The list of `n` number of cases retrieved from the query, where `n=page_size`.
    * **total_count**: `int` Total count of cases found by the query.
    """

    cases: List[Case]
    total_count: int = Field(alias="totalCount")


class QueryCasesRequest(Model):
    assignee: Optional[str] = None
    createdAt: Optional[str] = None
    isAssigned: Optional[bool] = None
    lastModifiedBy: Optional[str] = None
    name: Optional[str] = None
    pgNum: Optional[int] = 1
    pgSize: Optional[int] = 100
    srtDir: SortDirection = SortDirection.ASC
    srtKey: SortKeys = SortKeys.NUMBER
    status: Optional[CaseStatus] = None


class CreateCaseRequest(Model):
    name: str = Field(max_length=50)
    assignee: Optional[str] = None
    description: Optional[str] = Field(None, max_length=250)
    findings: Optional[str] = Field(None, max_length=30_000)
    subject: Optional[str] = None


class UpdateCaseRequest(Model, extra=Extra.ignore):
    name: Optional[str] = Field(
        None, description="The name of the case.", max_length=50
    )
    assignee: Optional[str] = None
    description: Optional[str] = Field(None, max_length=250)
    findings: Optional[str] = Field(None, max_length=30_000)
    subject: Optional[str] = None
    status: Optional[CaseStatus] = None


class RiskIndicator(BaseModel):
    name: str = None
    weight: int = None


class FileEvent(Model):
    event_id: Optional[str] = Field(
        None,
        alias="eventId",
        description="The unique identifier for the event.",
        examples=[
            "0_147e9445-2f30-4a91-8b2a-9455332e880a_973435567569502913_986467523038446097_163"
        ],
    )
    event_timestamp: Optional[datetime] = Field(
        None,
        alias="eventTimestamp",
        description="Date and time that the Code42 service on the device detected an event; based on the deviceâ€™s system clock and reported in Coordinated Universal Time (UTC).",
        examples=["2020-12-23T14:24:44.593Z"],
    )
    exposure: Optional[List[str]] = Field(
        None,
        description="Lists indicators that the data may be exposed.",
        examples=[["OutsideTrustedDomains", "IsPublic"]],
    )
    file_availability: Optional[Union[FileAvailability, str]] = Field(
        None,
        alias="fileAvailability",
        description="The download availability status of the file associated with the event.",
        examples=["EXACT_FILE_AVAILABLE"],
    )
    file_name: Optional[str] = Field(
        None,
        alias="fileName",
        description="The name of the file, including the file extension.",
        examples=["example.docx"],
    )
    file_path: Optional[str] = Field(
        None,
        alias="filePath",
        description="The file location on the user's device; a path forward or backslash should be included at the end of the filepath. Possibly null if the file event occurred on a cloud provider.",
        examples=["/Users/casey/Documents/"],
    )
    risk_indicators: Optional[List[Union[RiskIndicator, str]]] = Field(
        None,
        alias="riskIndicators",
        description="List of risk indicators identified for this event. If more than one risk indicator applies to this event, the sum of all indicators determines the total risk score.",
        table=lambda _list: list_as_panel([model_as_card(m) for m in _list])
        if len(_list)
        else "None",
    )
    risk_score: Optional[int] = Field(
        None,
        alias="riskScore",
        description="Sum of the weights for each risk indicator. This score is used to determine the overall risk severity of the event.",
        examples=[12],
    )
    risk_severity: Optional[str] = Field(
        None,
        alias="riskSeverity",
        description="The general risk assessment of the event, based on the numeric score.",
        examples=["CRITICAL"],
    )


class CaseFileEvents(ResponseModel):
    """A model representing file events associated with a case.

    **Fields**:

    * **events**: `List[FileEvent]` - List of events in the response..
    * **total_count**: `int` - Total number of events associated with the case.
    """

    events: Optional[List[FileEvent]] = Field(
        None, description="List of events in the response."
    )
    total_count: Optional[int] = Field(
        None,
        alias="totalCount",
        description="Total number of events associated with the case.",
        examples=[42],
    )



from __future__ import annotations

from datetime import datetime
from datetime import timedelta
from typing import List
from typing import Optional
from typing import Union

from isodate import duration_isoformat
from isodate import parse_duration
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import Field
from pydantic import model_validator
from pydantic import validate_arguments
from typing_extensions import Annotated

from _incydr_sdk.core.models import Model
from _incydr_sdk.enums.file_events import Category
from _incydr_sdk.enums.file_events import EventAction
from _incydr_sdk.enums.file_events import EventSearchTerm
from _incydr_sdk.enums.file_events import FileCategory
from _incydr_sdk.enums.file_events import Operator
from _incydr_sdk.enums.file_events import ReportType
from _incydr_sdk.enums.file_events import RiskIndicators
from _incydr_sdk.enums.file_events import RiskSeverity
from _incydr_sdk.enums.file_events import ShareType
from _incydr_sdk.enums.file_events import TrustReason
from _incydr_sdk.file_events.models.response import SavedSearch
from _incydr_sdk.file_events.models.response import SearchFilterGroup
from _incydr_sdk.file_events.models.response import SearchFilterGroupV2
from _incydr_sdk.queries.utils import parse_ts_to_date_str
from _incydr_sdk.queries.utils import parse_ts_to_ms_str

_term_enum_map = {
    "file.category": FileCategory,
    "event.action": EventAction,
    "source.category": Category,
    "destination.category": Category,
    "event.shareType": ShareType,
    "report.type": ReportType,
    "risk.indicators.name": RiskIndicators,
    "risk.severity": RiskSeverity,
    "risk.trustReason": TrustReason,
}


class Filter(BaseModel):
    term: str
    operator: Union[Operator, str]
    value: Optional[Union[int, str, List[str]]] = None
    model_config = ConfigDict(use_enum_values=True)

    @model_validator(mode="before")
    @classmethod
    def _validate_enums(cls, values: dict):  # noqa `root_validator` is a classmethod
        term = values.get("term")
        operator = values.get("operator")
        value = values.get("value")

        # 11-13-2024 - Removing strict filter term requirements to avoid breaking on new fields
        # make sure `term` is valid enum value
        # EventSearchTerm(term)

        if operator in (Operator.EXISTS, Operator.DOES_NOT_EXIST):
            values["value"] = None
            return values

        if operator in (Operator.IS, Operator.IS_NOT):
            if not isinstance(value, (str, int)):
                raise ValueError(
                    f"`IS` and `IS_NOT` filters require a `str | int` value, got term={term}, operator={operator}, value={value}."
                )

        # 11-13-2024 - Removing strict filter term requirements to avoid breaking on new fields
        # check that value is a valid enum for that search term
        # enum = _term_enum_map.get(term)
        # if enum:
        #     try:
        #         values.update(
        #             {"value": enum[value.upper()]}
        #         )  # check if enum name is passed as a value
        #     except KeyError:
        #         enum(value)

        return values


class FilterGroup(BaseModel):
    filterClause: str = "AND"
    filters: Optional[List[Filter]] = None


class FilterGroupV2(BaseModel):
    subgroupClause: str = "AND"
    subgroups: List[Union[FilterGroupV2, FilterGroup]]


class Query(Model):
    groupClause: str = "AND"
    groups: Optional[List[FilterGroup]]
    pgNum: int = 1
    pgSize: int = 100
    pgToken: Optional[str]
    srtDir: str = "asc"
    srtKey: EventSearchTerm = "event.id"


class EventQuery(Model):
    """
    Class to build a file event query. Use the class methods to attach additional filter operators.

    **Parameters**:

         * **start_date**: `int`, `float`, `str`, `datetime`, `timedelta` -  Start of the date range to query for events. Defaults to None.
         * **end_date**: `int`, `float`, `str`, `datetime` - End of the date range to query for events.  Defaults to None.
    """

    group_clause: str = Field("AND", alias="groupClause")
    groups: Optional[List[FilterGroup]]
    page_num: int = Field(1, alias="pgNum")
    page_size: Annotated[int, Field(le=10000)] = Field(100, alias="pgSize")
    page_token: Optional[str] = Field("", alias="pgToken")
    sort_dir: str = Field("asc", alias="srtDir")
    sort_key: EventSearchTerm = Field("event.id", alias="srtKey")
    model_config = ConfigDict(
        validate_assignment=True,
        use_enum_values=True,
        json_encoders={datetime: lambda dt: dt.isoformat().replace("+00:00", "Z")},
    )

    def __init__(
        self,
        start_date: Union[datetime, timedelta, int, float, str] = None,
        end_date: Union[datetime, int, float, str] = None,
        **kwargs,
    ):
        groups = kwargs.get("groups") or []

        if start_date or end_date:
            groups.append(_create_date_range_filter_group(start_date, end_date))

        kwargs["groups"] = groups
        super().__init__(**kwargs)

    def equals(self, term: str, values: Union[str, List[str]]):
        """
        Adds an `equals` filter to the query.

        When passed as part of a query, returns events when the field corresponding to the filter term equals the
        indicated value(s).

        Example:
            `EventQuery(**kwargs).equals('file.category', 'Document')` creates a query which will return file events
            where the `file.category` field is equal to `Document`.

        **Parameters**:

        * **term**: `str` - The term which corresponds to a file event field.
        * **values**: `str | List[str]` - The value(s) for the term to match.
        """
        if isinstance(values, str):
            values = [values]
        if len(values) < 1:
            raise ValueError("equals() requires at least one value.")
        filters = [Filter(term=term, operator=Operator.IS, value=val) for val in values]
        filter_group = FilterGroup(
            filters=filters,
            filterClause="OR" if len(values) > 1 else "AND",
        )
        self.groups.append(filter_group)
        return self

    def not_equals(self, term, values: Union[str, List[str]]):
        """
        Adds an `not_equals` filter to the query. The opposite of the `equals` filter.

        When passed as part of a query, returns events when the field corresponding to the filter term does not equal the indicated value(s).

        Example:
            `EventQuery(**kwargs).not_equals('file.category', 'Document')` creates a query which will return file events where the `file.category` field is NOT equal to `Document`.

        **Parameters**:

        * **term**: `str` - The term which corresponds to a file event field.
        * **values**: `str | List[str]` - The value(s) for the term to not match.
        """

        if isinstance(values, str):
            values = [values]
        if len(values) < 1:
            raise ValueError("not_equals() requires at least one value.")
        filters = [
            Filter(term=term, operator=Operator.IS_NOT, value=val) for val in values
        ]
        filter_group = FilterGroup(
            filters=filters,
            filterClause="AND",
        )
        self.groups.append(filter_group)
        return self

    def exists(self, term: str):
        """
        Adds an `exists` filter to the query. The opposite of the `does_not_exist` filter.

        When passed as part of a query, returns events when the field corresponding to the filter term is not `null`.

        Example:
            `EventQuery(**kwargs).exists('risk.trustReason')` creates a query which will return file events where the `risk.trustReason` field is populated with any not null value.

        **Parameters**:

        * **term**: `str` - The term which corresponds to a file event field.
        """
        self.groups.append(
            FilterGroup(filters=[Filter(term=term, operator=Operator.EXISTS)])
        )
        return self

    def does_not_exist(self, term: str):
        """
        Adds a `does_not_exist` filter to the query.

        When passed as part of a query, returns events when the field corresponding to the filter term is `null`.

        Example:
            `EventQuery(**kwargs).does_not_exist('risk.TrustReason')` creates a query which will return file events where the `risk.trustReason` field is null.

        **Parameters**:

        * **term**: `str` - The term which corresponds to a file event field.
        """
        self.groups.append(
            FilterGroup(filters=[Filter(term=term, operator=Operator.DOES_NOT_EXIST)])
        )
        return self

    @validate_arguments
    def greater_than(self, term: str, value: int):
        """
        Adds a `greater_than` filter to the query. The opposite of the `less_than` filter.

        When passed as part of a query, returns events when the field corresponding to the filter term is greater than the indicated value.

        Example:
            `EventQuery(**kwargs).greater_than('risk.score', 10)` creates a query which will return file events where the `risk.score` field is greater than `10`.

        **Parameters**:

        * **term**: `str` - The term which corresponds to a file event field.
        * **values**: `int` - The value for the term to be greater than.
        """

        self.groups.append(
            FilterGroup(
                filters=[Filter(term=term, operator=Operator.GREATER_THAN, value=value)]
            )
        )
        return self

    @validate_arguments
    def less_than(self, term: str, value: int):
        """
        Adds a `less_thn` filter to the query. The opposite of the `greater_than` filter.

        When passed as part of a query, returns events when the field corresponding to the filter term is less than the indicated value.

        Example:
            `EventQuery(**kwargs).less_than('risk.score', 10)` creates a query which will return file events where the `risk.score` field is less than `10`.

        **Parameters**:

        * **term**: `str` - The term which corresponds to a file event field.
        * **values**: `int` - The value for the term to be less than.
        """
        self.groups.append(
            FilterGroup(
                filters=[Filter(term=term, operator=Operator.LESS_THAN, value=value)]
            )
        )
        return self

    @validate_arguments
    def is_any(self, term: str, values: List[str]):
        """
        Adds a `is_any` filter to the query. The opposite of the `is_none` filter.

        When passed as part of a query, returns events when the field corresponding to the filter term matches any of the provided values.

        Example:
            `EventQuery(**kwargs).is_any("destination.category", ["AI Tools", "Cloud Storage"])` creates a query which will return file events where the destination category is either AI Tools or Cloud Storage.

        **Parameters**:

        * **term**: `str` - The term which corresponds to a file event field.
        * **values**: `List[str]` - The values to match.
        """
        self.groups.append(
            FilterGroup(
                filters=[Filter(term=term, operator=Operator.IS_ANY, value=values)]
            )
        )
        return self

    @validate_arguments
    def is_none(self, term: str, values: List[str]):
        """
        Adds a `is_none` filter to the query. The opposite of the `is_any` filter.

        When passed as part of a query, returns events when the field corresponding to the filter term matches none of the provided values.

        Example:
            `EventQuery(**kwargs).is_any("destination.category", ["AI Tools", "Cloud Storage"])` creates a query which will return file events where the destination category is anything other than AI Tools or Cloud Storage.

        **Parameters**:

        * **term**: `str` - The term which corresponds to a file event field.
        * **values**: `List[str]` - The values for the term to not match.
        """
        self.groups.append(
            FilterGroup(
                filters=[Filter(term=term, operator=Operator.IS_NONE, value=values)]
            )
        )
        return self

    def date_range(self, term: str, start_date=None, end_date=None):
        """
        Adds a date-based filter for the specified term.

        When passed as part of a query, returns events within the specified date range, or all events before/after the specified date if only one of start_date or end_date is given.

        Example:
            `EventQuery(**kwargs).date_range(term="event.inserted", start_date="P1D")` creates a query that returns all events inserted into Forensic Search within the past day.

        **Parameters**:

        * **term**: `str` - The term which corresponds to a file event field.
        * **start_date**: `int`, `float`, `str`, `datetime`, `timedelta` -  Start of the date range to query for events. Defaults to None.
        * **end_date**: `int`, `float`, `str`, `datetime` - End of the date range to query for events.  Defaults to None.
        """
        if start_date or end_date:
            self.groups.append(
                _create_date_range_filter_group(
                    start_date=start_date, end_date=end_date, term=term
                )
            )
        return self

    def on(self, term: str, date=None):
        """
        Adds a date-based filter for the specified term.

        When passed as part of a query, returns events on the specified date.

        Example:
            `EventQuery(**kwargs).date_range(term="event.inserted", start_date="P1D")` creates a query that returns all events inserted into Forensic Search within the past day.

        **Parameters**:

        * **term**: `str` - The term which corresponds to a file event field.
        * **date**: `int`, `float`, `str`, `datetime` -  The date to query for events. Defaults to None.
        """
        self.groups.append(
            FilterGroup(
                filters=[
                    Filter(
                        term=term,
                        operator=Operator.ON,
                        value=parse_ts_to_date_str(date),
                    )
                ]
            )
        )
        return self

    def matches_any(self):
        """
        Sets operator to combine multiple filters to `OR`.
        Returns events that match at least one of the filters in the query.

        Default operator is `AND`, which returns events that match all filters in the query.
        """
        self.group_clause = "OR"
        return self

    def subquery(self, subgroup_query: EventQuery):
        """
        Adds a subgroup to the query, with any filter groups or subgroups from the subgroup_query added to the present query.

        Example:
            `EventQuery().greater_than("risk.score", 1).subquery(EventQuery().matches_any().equals("destination.category", "AI Tools").equals("file.name", "example"))`

        This example creates a query which matches events having a risk score of 1 or greater and have a destination category equal to "AI Tools" or have a filename equal to "example"

        **Parameters**:

        * **subgroup_query**: `EventQuery` - An EventQuery object. The filter groups and subgroups will be added to the present query. The subgroup query's group clause will be used for the created subgroup.
        """
        self.groups.append(
            FilterGroupV2(
                subgroupClause=subgroup_query.group_clause,
                subgroups=subgroup_query.groups,
            )
        )
        return self

    @classmethod
    def from_saved_search(cls, saved_search: SavedSearch):
        """
        Create an `EventQuery` object from a `SavedSearch` response.
        """
        query = cls()
        if saved_search.group_clause:
            query.group_clause = saved_search.group_clause
        if saved_search.groups:
            for i in saved_search.groups:
                query.groups.append(_handle_filter_group_type(i))
        if saved_search.srt_dir:
            query.sort_dir = saved_search.srt_dir
        if saved_search.srt_key:
            query.sort_key = saved_search.srt_key
        return query


def _create_date_range_filter_group(start_date, end_date, term=None):
    def _validate_duration_str(iso_duration_str):
        try:
            parse_duration(iso_duration_str)
        except Exception:
            return False
        return True

    filters = []

    if isinstance(start_date, timedelta) or _validate_duration_str(start_date):
        if isinstance(start_date, timedelta):
            start_date = duration_isoformat(start_date)
        filters.append(
            Filter(
                term=term or EventSearchTerm.TIMESTAMP,
                operator=Operator.WITHIN_THE_LAST,
                value=start_date,
            )
        )
    else:
        if start_date:
            filters.append(
                Filter(
                    term=term or EventSearchTerm.TIMESTAMP,
                    operator=Operator.ON_OR_AFTER,
                    value=parse_ts_to_ms_str(start_date),
                )
            )

        if end_date:
            filters.append(
                Filter(
                    term=term or EventSearchTerm.TIMESTAMP,
                    operator=Operator.ON_OR_BEFORE,
                    value=parse_ts_to_ms_str(end_date),
                )
            )
    return FilterGroup(filters=filters)


def _create_filter_group(filter_group: SearchFilterGroup) -> FilterGroup:
    filters = [
        Filter.model_construct(value=f.value, operator=f.operator, term=f.term)
        for f in filter_group.filters
    ]
    return FilterGroup.model_construct(
        filterClause=filter_group.filter_clause, filters=filters
    )


def _create_filter_group_v2(filter_group_v2: SearchFilterGroupV2) -> FilterGroupV2:
    subgroups = []
    for subgroup in filter_group_v2.subgroups:
        subgroups.append(_handle_filter_group_type(subgroup))
    return FilterGroupV2.model_construct(
        subgroupClause=filter_group_v2.subgroup_clause, subgroups=subgroups
    )


def _handle_filter_group_type(
    filter_group: Union[SearchFilterGroup, SearchFilterGroupV2]
) -> Union[FilterGroup, FilterGroupV2]:
    if isinstance(filter_group, SearchFilterGroup):
        return _create_filter_group(filter_group)
    if isinstance(filter_group, SearchFilterGroupV2):
        return _create_filter_group_v2(filter_group)
    else:
        raise TypeError(
            "Query filter group must be one of: SearchFilterGroup, SearchFilterGroupV2"
        )



from __future__ import annotations

import json
from datetime import datetime
from typing import List
from typing import Optional

import rich.box
from pydantic import Field
from pydantic import field_validator
from pydantic import StringConstraints
from rich.markdown import Markdown
from rich.panel import Panel
from typing_extensions import Annotated

from _incydr_sdk.core.models import Model


class Observation(Model):
    id: Optional[str] = Field(
        None, description="Id of given observation.", examples=["uniqueObservationId"]
    )
    observed_at: datetime = Field(
        None,
        alias="observedAt",
        description="Timestamp when the activity was first observed.",
        examples=["2020-02-19T01:57:45.006683Z"],
    )
    last_observed_at: Optional[datetime] = Field(
        None,
        alias="lastObservedAt",
        description="Timestamp when the activity was last observed.",
        examples=["2020-02-19T01:57:45.006683Z"],
    )
    type: Optional[str] = Field(
        None,
        description="The type of observation data recorded.",
        examples=["FedCloudSharePermissions"],
    )
    data: Optional[dict] = Field(
        None,
        description="The JSON formatted observation data rolled into one aggregation.",
        examples=[
            '{"type$":"OBSERVED_CLOUD_SHARE_ACTIVITY","id":"exampleId","sources":["OneDrive"],"exposureTypes":["PublicLinkShare"],"firstActivityAt":"2020-02-19T01:50:00.0000000Z","lastActivityAt":"2020-02-19T01:55:00.0000000Z","fileCount":2,"totalFileSize":200,"fileCategories":[{"type$":"OBSERVED_FILE_CATEGORY","category":"Document","fileCount":2,"totalFileSize":53,"isSignificant":false}],"outsideTrustedDomainsEmailsCount":0,"outsideTrustedDomainsTotalDomainCount":0,"outsideTrustedDomainsTotalDomainCountTruncated":false}'
        ],
        table=lambda data: json.dumps(data, indent=2),
    )

    @field_validator("data", mode="before")
    @classmethod
    def parse_data_json(cls, value):  # noqa
        return json.loads(value) if value else None


class Note(Model):
    id: Optional[str] = Field(
        None, description="Unique id of the note.", examples=["noteId"]
    )
    last_modified_at: datetime = Field(
        None,
        alias="lastModifiedAt",
        description="Timestamp of when the note was last modified.",
        examples=["2020-02-19T01:57:45.006683Z"],
    )
    last_modified_by: Optional[str] = Field(
        None,
        alias="lastModifiedBy",
        description="User who last modified the note.",
        examples=["exampleUsername"],
    )
    message: Optional[str] = Field(
        None,
        description="The note itself.",
        examples=["This is a note."],
        table=lambda note: Panel(Markdown(note), width=120, box=rich.box.MINIMAL)
        if note
        else note,
    )


class AuditInfo(Model):
    modified_by: Optional[str] = Field(
        None,
        alias="modifiedBy",
        description="Username of the individual who last modified the rule.",
        examples=["UserWhoMostRecentlyModifiedTheRule"],
    )
    modified_at: datetime = Field(
        None,
        alias="modifiedAt",
        description="Timestamp of when the rule was last modified.",
        examples=["2020-02-19T01:57:45.006683Z"],
    )


class Watchlist(Model):
    id: Optional[str] = Field(
        None, description="Unique id of this watchlist.", examples=["guid"]
    )
    name: Optional[str] = Field(
        None, description="Name of the watchlist.", examples=["Development Department"]
    )
    type: Optional[str] = Field(
        None, description="Type of watchlist.", examples=["DEPARTING_EMPLOYEE"]
    )
    is_significant: bool = Field(
        None,
        alias="isSignificant",
        description="Indicates whether the watchlist was part of the triggering rule's criteria.",
        examples=["true"],
    )


class ObserverRuleMetadata(AuditInfo):
    name: Optional[str] = Field(
        None,
        description="The name of the rule.",
        examples=["My Removable Media Exfiltration Rule"],
    )
    description: Optional[str] = Field(
        None,
        description="The description of the rule.",
        examples=["Will generate alerts when files moved to USB."],
    )
    severity: Optional[str] = Field(
        None, description="The static severity of the rule (deprecated)."
    )
    is_system: Optional[bool] = Field(
        None,
        alias="isSystem",
        description="Boolean indicating if the rule was created from another Code42 Application.",
        examples=["FALSE"],
    )
    is_enabled: bool = Field(
        None,
        alias="isEnabled",
        description="Boolean indicating if the rule is enabled to trigger alerts.",
        examples=["TRUE"],
    )
    rule_source: Optional[str] = Field(
        None,
        alias="ruleSource",
        description="The source of the rule.  Will be one of [DepartingEmployee, Alerting, HighRiskEmployee]",
        examples=["Alerting"],
    )


class AlertSummary(Model):
    """
    A model representing an alert summary.

    **Fields**:

    * **tenant_id**: `str` The unique identifier representing the tenant.
    * **type**: `RuleType` Rule type that generated the alert.
    * **id**: `str` The unique id of the alert.
    * **created_at**: `datetime` The timestamp when the alert was created.
    * **state**: [`AlertState`][alert-state] The current state of the alert.
    * **state_last_modified_by**: `str` The actor who last modified the alert state.
    * **state_last_modified_at**: `datetime` The timestamp when the alert state was last modified.
    * **name**: `str` The name of the alert.  Same as the name of the rule that triggered it.
    * **description**: `str` The description of the alert.  Same as the description of the rule that triggered it.
    * **actor**: `str` The user who triggered the alert.
    * **actor_id**: `str` The user id who triggered the alert, if it is available.
    * **target**: `str` Unused legacy property.
    * **severity**: [`AlertSeverity`][alert-severity] Indicates static rule severity of the alert. (Deprecated)
    * **risk_severity**: [`RiskSeverity`][risk-severity] Indicates event risk severity of the alert.
    * **rule_id**: `str` The unique id corresponding to the rule which triggered the alert.
    * **watchlists**: `str` Watchlists the actor is on at the time of the alert (if any).
    """

    tenant_id: Annotated[str, StringConstraints(max_length=40)] = Field(
        None,
        alias="tenantId",
        description="The unique identifier representing the tenant.",
        examples=["MyExampleTenant"],
    )
    type: Optional[str] = Field(..., description="Rule type that generated the alert.")
    id: Optional[str] = Field(
        None, description="The unique id of the alert.", examples=["alertId"]
    )
    created_at: datetime = Field(
        None,
        alias="createdAt",
        description="The timestamp when the alert was created.",
        examples=["2020-02-19T01:57:45.006683Z"],
    )
    state: Optional[str] = Field(..., description="The current state of the alert.")
    state_last_modified_by: Optional[str] = Field(None, alias="stateLastModifiedBy")
    state_last_modified_at: Optional[datetime] = Field(
        None, alias="stateLastModifiedAt"
    )
    name: Optional[str] = Field(
        None,
        description="The name of the alert.  Same as the name of the rule that triggered it.",
        examples=["Removable Media Exfiltration Rule"],
    )
    description: Optional[str] = Field(
        None,
        description="The description of the alert.  Same as the description of the rule that triggered it.",
        examples=["Alert me on all removable media exfiltration."],
    )
    actor: Optional[str] = Field(
        None,
        description="The user who triggered the alert.",
        examples=["exampleUser@mycompany.com"],
    )
    actor_id: Optional[str] = Field(
        None,
        alias="actorId",
        description="The authority user id who triggered the alert, if it is available.",
        examples=["authorityUserId"],
    )
    target: Optional[str] = None
    severity: Optional[str] = Field(
        None, description="Indicates static rule severity of the alert."
    )
    risk_severity: Optional[str] = Field(
        None,
        alias="riskSeverity",
        description="Indicates event risk severity of the alert.",
        examples=["MODERATE"],
    )
    rule_id: Optional[str] = Field(
        None,
        alias="ruleId",
        description="The unique id corresponding to the rule which triggered the alert.",
        examples=["uniqueRuleId"],
    )
    watchlists: Optional[List[Watchlist]] = Field(
        None,
        description="Watchlists the actor is on at the time of the alert.",
        examples=[[]],
        table=lambda watchlists: "\n".join((w.name or w.type) for w in watchlists)
        if watchlists
        else watchlists,
        csv=lambda watchlists: "|".join((w.name or w.type) for w in watchlists)
        if watchlists
        else watchlists,
    )


class ObserverRuleMetadataEssentials(ObserverRuleMetadata):
    tenant_id: Annotated[str, StringConstraints(max_length=40)] = Field(
        None,
        alias="tenantId",
        description="The unique identifier representing the tenant.",
        examples=["MyExampleTenant"],
    )
    observer_rule_id: Optional[str] = Field(
        None,
        alias="observerRuleId",
        description="Id of the rule in the observer.",
        examples=["UniqueRuleId"],
    )
    type: Optional[str] = Field(None, description="Rule type of the rule.")


class AlertDetails(AlertSummary):
    """
    A model representing the full details of an alert. Includes all the fields from `AlertSummary` plus file event
    observations (the events that triggered the alert), and any notes that have been added to the alert.

    **Fields**:

    * **observations**: `List[Observation]` List of observed file events that triggered the alert.
    * **note**: `str` Most recent note added to the alert.
    """

    observations: Optional[List[Observation]] = None
    note: Optional[Note] = None



from __future__ import annotations

from typing import List
from typing import Optional

from pydantic import BaseModel
from pydantic import Field
from pydantic import StringConstraints
from typing_extensions import Annotated

from _incydr_sdk.enums.alerts import AlertState


class UpdateAlertStateRequest(BaseModel):
    tenant_id: Annotated[str, StringConstraints(max_length=40)] = Field(
        ...,
        alias="tenantId",
        description="The unique identifier representing the tenant.",
        examples=["MyExampleTenant"],
    )
    alert_ids: List[str] = Field(
        ...,
        alias="alertIds",
        description="The unique identifiers representing the alerts you want to act upon.",
        examples=[["ExampleAlertId1", "ExampleAlertId2"]],
        max_length=100,
    )
    state: AlertState = Field(
        ..., description="The state to update the given alerts to."
    )
    note: Optional[Annotated[str, StringConstraints(max_length=2000)]] = Field(
        None,
        description="An optional note to attach to the alert",
        examples=["This is an example note."],
    )


class AlertDetailsRequest(BaseModel):
    alert_ids: List[str] = Field(
        ...,
        alias="alertIds",
        description="The unique identifiers representing the alerts you want to act upon.",
        examples=[["ExampleAlertId1", "ExampleAlertId2"]],
        max_length=100,
    )


class AddNoteRequest(BaseModel):
    tenant_id: Annotated[str, StringConstraints(max_length=40)] = Field(
        ...,
        alias="tenantId",
        description="The unique identifier representing the tenant.",
        examples=["MyExampleTenant"],
    )
    alert_id: Annotated[str, StringConstraints(max_length=40)] = Field(
        ...,
        alias="alertId",
        description="The unique identifier representing the alert you want to act upon.",
        examples=["ExampleAlertId"],
    )
    note: Annotated[str, StringConstraints(max_length=2000)] = Field(
        ...,
        description="The note to attach to the alert.",
        examples=["This is an example note."],
    )



import json
import os
from os import path

import click


class Cursor:
    def __init__(self, location):
        self._location = location
        self._name = path.basename(location)

    @property
    def name(self):
        return self._name

    @property
    def value(self):
        with open(self._location) as f:
            return f.read()


class CursorStore:
    """
    Cursor store for checkpoints.

    Properties:
        * dir_path - directory where files will be saved.
        * event_key - the string to be suffixed to the end of files which are storing events/event hashes for a checkpoint.

            Example:
                the file `chckpt_audit_events` is storing hashes of audit log events for the checkpoint called 'chcktpt'
    """

    def __init__(self, dir_path, event_key):
        self._dir_path = dir_path
        self._event_key = event_key

    def get(self, cursor_name):
        """Gets the last stored date observed timestamp."""
        try:
            location = path.join(self._dir_path, cursor_name)
            with open(location) as f:
                c = f.read()
                return c.strip() if c else None
        except FileNotFoundError:
            return None

    def replace(self, cursor_name, new_checkpoint):
        """Replaces the last stored date observed timestamp with the given one."""
        location = path.join(self._dir_path, cursor_name)
        with open(location, "w") as checkpoint:
            checkpoint.write(str(new_checkpoint))

    def delete(self, cursor_name):
        """Removes a single cursor from the store."""
        try:
            location = path.join(self._dir_path, cursor_name)
            os.remove(location)
        except FileNotFoundError:
            msg = f"No checkpoint named {cursor_name} found for this API client."
            raise click.BadOptionUsage("checkpoint", msg)

    def clean(self):
        """Removes all cursors from this store."""
        cursors = self.get_all_cursors()
        for cursor in cursors:
            self.delete(cursor.name)

    def get_all_cursors(self):
        """Returns a list of all cursors stored in this directory (which is typically scoped to a profile)."""
        dir_contents = os.listdir(self._dir_path)
        return [Cursor(f) for f in dir_contents if self._is_file(f)]

    def _is_file(self, node_name):
        return path.isfile(path.join(self._dir_path, node_name))

    def get_items(self, cursor_name):
        """
        Used with alerts, audit_log events, and sessions to avoid duplicates
        """
        try:
            location = path.join(self._dir_path, cursor_name) + f"_{self._event_key}"
            with open(location) as checkpoint:
                return json.loads(checkpoint.read())
        except (FileNotFoundError, json.JSONDecodeError):
            return []

    def replace_items(self, cursor_name, new_events):
        """
        Used with alerts, audit_log events, and sessions to avoid duplicates
        """
        location = path.join(self._dir_path, cursor_name) + f"_{self._event_key}"
        with open(location, "w") as checkpoint:
            checkpoint.write(json.dumps(new_events))



from . import agents
from . import alerts
from . import cases
from . import devices
from . import file_events
from . import trusted_activities
from . import watchlists
from _incydr_sdk.enums import SortDirection  # noqa



from typing import Optional

import click
from rich.panel import Panel

from _incydr_cli import console
from _incydr_cli import logging_options
from _incydr_cli import render
from _incydr_cli.cmds.options.output_options import columns_option
from _incydr_cli.cmds.options.output_options import single_format_option
from _incydr_cli.cmds.options.output_options import SingleFormat
from _incydr_cli.cmds.options.output_options import table_format_option
from _incydr_cli.cmds.options.output_options import TableFormat
from _incydr_cli.core import IncydrCommand
from _incydr_cli.core import IncydrGroup
from _incydr_sdk.core.client import Client
from _incydr_sdk.orgs.models import Org
from _incydr_sdk.utils import model_as_card


@click.group(cls=IncydrGroup)
@logging_options
def orgs():
    """View and manage orgs."""


@orgs.command("activate", cls=IncydrCommand)
@click.argument("org_guid")
@logging_options
def activate_org(org_guid: str):
    """
    Activate the given org.
    """
    client = Client()
    client.orgs.v1.activate(org_guid)
    console.print(f"Org '{org_guid}' successfully activated.")


@orgs.command("list", cls=IncydrCommand)
@click.option(
    "--active/--inactive",
    default=None,
    help="Filter by active or inactive orgs. Defaults to returning both when when neither option is passed.",
)
@table_format_option
@columns_option
@logging_options
def list_(
    active: Optional[bool],
    format_: TableFormat,
    columns: Optional[str],
):
    """
    List orgs.
    """
    client = Client()
    orgs_ = client.orgs.v1.list(active=active).orgs

    if format_ == TableFormat.csv:
        render.csv(Org, orgs_, columns=columns, flat=True)
    elif format_ == TableFormat.table:
        render.table(Org, orgs_, columns=columns, flat=False)
    elif format_ == TableFormat.json_pretty:
        for item in orgs_:
            console.print_json(item.json())
    else:
        for item in orgs_:
            click.echo(item.json())


@orgs.command("create", cls=IncydrCommand)
@click.argument("name")
@click.option(
    "--external-reference",
    default=None,
    help="The external reference string for the org. Defaults to None.",
)
@click.option(
    "--notes",
    default=None,
    help="The notes string for the org. Defaults to None.",
)
@click.option(
    "--parent-org-guid",
    default=None,
    help="The org guid for the created org's parent. Defaults to your tenant's parent org.",
)
@single_format_option
@columns_option
@logging_options
def create(
    name: str,
    external_reference: Optional[str],
    notes: Optional[str],
    parent_org_guid: Optional[str],
    format_: SingleFormat,
    columns: Optional[str],
):
    """
    Create a new org.
    """
    client = Client()
    org = client.orgs.v1.create(
        org_name=name,
        org_ext_ref=external_reference,
        parent_org_guid=parent_org_guid,
        notes=notes,
    )

    if format_ == SingleFormat.rich and client.settings.use_rich:
        console.print(Panel.fit(model_as_card(org), title=f"Org {org.org_name}"))
    elif format_ == SingleFormat.json_pretty:
        console.print_json(org.json())
    else:
        click.echo(org.json())


@orgs.command("deactivate", cls=IncydrCommand)
@click.argument("org_guid")
@logging_options
def deactivate_org(org_guid: str):
    """
    Deactivate the given org.
    """
    client = Client()
    client.orgs.v1.deactivate(org_guid)
    console.print(f"Org '{org_guid}' successfully deactivated.")


@orgs.command("show", cls=IncydrCommand)
@click.argument("org_guid")
@single_format_option
@columns_option
@logging_options
def show(
    org_guid: str,
    format_: SingleFormat,
    columns: Optional[str],
):
    """
    View details of an org.
    """
    client = Client()
    org = client.orgs.v1.get_org(org_guid=org_guid)

    if format_ == SingleFormat.rich and client.settings.use_rich:
        console.print(Panel.fit(model_as_card(org), title=f"Org: {org.org_name}"))
    elif format_ == SingleFormat.json_pretty:
        console.print_json(org.json())
    else:
        click.echo(org.json())


@orgs.command("update", cls=IncydrCommand)
@click.argument("org_guid")
@click.option(
    "--name", help="The name of the org being updated. Required.", required=True
)
@click.option(
    "--external-reference",
    default=None,
    help="The external reference string for the org. Defaults to None.",
)
@click.option(
    "--notes",
    default=None,
    help="The notes string for the org. Defaults to None.",
)
@single_format_option
@columns_option
@logging_options
def update(
    org_guid: str,
    name: str,
    external_reference: Optional[str],
    notes: Optional[str],
    format_: SingleFormat,
    columns: Optional[str],
):
    """
    Update an org.
    """
    client = Client()
    org = client.orgs.v1.update(
        org_guid=org_guid, org_name=name, org_ext_ref=external_reference, notes=notes
    )

    if format_ == SingleFormat.rich and client.settings.use_rich:
        console.print(Panel.fit(model_as_card(org), title=f"Org {org.org_name}"))
    elif format_ == SingleFormat.json_pretty:
        console.print_json(org.json())
    else:
        click.echo(org.json())



import base64
import json
import logging
from collections import deque

from requests_toolbelt import user_agent
from requests_toolbelt.sessions import BaseUrlSession

from _incydr_sdk.__version__ import __version__
from _incydr_sdk.actors.client import ActorsClient
from _incydr_sdk.agents.client import AgentsClient
from _incydr_sdk.alert_rules.client import AlertRulesClient
from _incydr_sdk.alerts.client import AlertsClient
from _incydr_sdk.audit_log.client import AuditLogClient
from _incydr_sdk.cases.client import CasesClient
from _incydr_sdk.core.auth import APIClientAuth
from _incydr_sdk.core.auth import RefreshTokenAuth
from _incydr_sdk.core.settings import IncydrSettings
from _incydr_sdk.customer.client import CustomerClient
from _incydr_sdk.departments.client import DepartmentsClient
from _incydr_sdk.devices.client import DevicesClient
from _incydr_sdk.directory_groups.client import DirectoryGroupsClient
from _incydr_sdk.file_events.client import FileEventsClient
from _incydr_sdk.files.client import FilesClient
from _incydr_sdk.legal_hold.client import LegalHoldClient
from _incydr_sdk.orgs.client import OrgsClient
from _incydr_sdk.risk_profiles.client import RiskProfiles
from _incydr_sdk.sessions.client import SessionsClient
from _incydr_sdk.trusted_activities.client import TrustedActivitiesClient
from _incydr_sdk.users.client import UsersClient
from _incydr_sdk.watchlists.client import WatchlistsClient

_base_user_agent = user_agent("incydrSDK", __version__)


class Client:
    """
    An HTTP client for interacting with the Code42 Incydr API.

    **Parameters**:

    * **url**: `str` The url of your Incydr API gateway. See the [developer getting started guide](https://developer.code42.com/api/#section/User-Guides/Get-started)
        to find your API domain based on your console login URL.
    * **api_client_id**: `str` The ID of your [Incydr API Client](https://support.code42.com/hc/en-us/articles/14827617150231)
    * **api_client_secret**: `str` The Secret for your Incydr API Client.

    Usage example:

        >>> import incydr
        >>> client = incydr.Client(url="<api_domain>", api_client_id="<client_id>", api_client_secret="<client_secret>")

    """

    def __init__(
        self,
        url: str = None,
        api_client_id: str = None,
        api_client_secret: str = None,
        skip_auth: bool = False,
        **settings_kwargs,
    ):
        self._settings = IncydrSettings(
            url=url,
            api_client_id=api_client_id,
            api_client_secret=api_client_secret,
            **settings_kwargs,
        )
        self._request_history = deque(maxlen=self._settings.max_response_history)

        self._session = BaseUrlSession(base_url=self._settings.url)
        self._session.headers["User-Agent"] = (
            self._settings.user_agent_prefix or ""
        ) + _base_user_agent
        if self._settings.refresh_token and self._settings.refresh_url:
            self._session.auth = RefreshTokenAuth(
                session=self._session,
                refresh_url=self._settings.refresh_url,
                refresh_token=self._settings.refresh_token.get_secret_value(),
            )
        else:
            self._session.auth = APIClientAuth(
                session=self._session,
                api_client_id=self._settings.api_client_id,
                api_client_secret=self._settings.api_client_secret,
            )

        def response_hook(response, *args, **kwargs):
            level = self._settings.log_level
            if level == logging.INFO:
                self.settings._log_response_info(response)
            if level == logging.DEBUG:
                self.settings._log_response_debug(response)

            self._request_history.appendleft(response)
            response.raise_for_status()

        self._session.hooks["response"] = [response_hook]

        self._actors = ActorsClient(self)
        self._agents = AgentsClient(self)
        self._alerts = AlertsClient(self)
        self._alert_rules = AlertRulesClient(self)
        self._audit_log = AuditLogClient(self)
        self._cases = CasesClient(self)
        self._customer = CustomerClient(self)
        self._departments = DepartmentsClient(self)
        self._devices = DevicesClient(self)
        self._directory_groups = DirectoryGroupsClient(self)
        self._file_events = FileEventsClient(self)
        self._files = FilesClient(self)
        self._legal_hold = LegalHoldClient(self)
        self._orgs = OrgsClient(self)
        self._sessions = SessionsClient(self)
        self._trusted_activities = TrustedActivitiesClient(self)
        self._users = UsersClient(self)
        self._risk_profiles = RiskProfiles(self)
        self._watchlists = WatchlistsClient(self)

        if not skip_auth:
            self._session.auth.refresh()

    @property
    def tenant_id(self):
        """Property returning the current tenant ID."""
        token = self.session.auth.token_response.access_token.get_secret_value()
        payload = token.encode("ascii").split(b".")[-2]
        extra = len(payload) % 4
        if extra > 0:
            payload += b"=" * (4 - extra)
        return json.loads(base64.urlsafe_b64decode(payload))["tenantUid"]

    @property
    def request_history(self):
        """
        Property returning a list of the last `n` number of [`requests.Response`](https://requests.readthedocs.io/en/latest/api/#requests.Response)
        objects received, where `n` equals the `client.settings.max_response_history` value (default=5).

        The most recent request is the first item in the list: `client.request_history[0]`
        """
        return list(self._request_history)

    @property
    def settings(self):
        """
        Property returning an [`IncydrSettings`](../settings) object that contains the configuration for this client.
        """
        return self._settings

    @property
    def session(self):
        """
        Property returning the core [`requests.Session`](https://requests.readthedocs.io/en/latest/api/#request-sessions) used to make all
        HTTP requests.

        Contains a hook that prepends the base url of the Code42 API gateway to each request url.

        Usage:

            >>> response = client.session.get("/v1/users")
            >>> response
            <Response [200]>
            >>> response.url
            'https://api.us.code42.com/v1/users'
        """
        return self._session

    @property
    def actors(self):
        """
        Property returning an [`ActorsClient`](../actors) for interacting with `/v*/actors` API endpoints.

        Usage:
            >>> clients.actors.v1.get_page()
        """
        return self._actors

    @property
    def agents(self):
        """
        Property returning an [`AgentsClient`](../agents) for interacting with
        `/v*/agents` API endpoints.
        Usage:
            >>> client.agents.v1.get_page()
        """
        return self._agents

    @property
    def alerts(self):
        """
        (DEPRECATED) Property returning an [`AlertsClient`](../alerts) for interacting with
        `/v*/alerts` API endpoints.
        Usage:
            >>> client.alerts.v1.get_page()
        """
        return self._alerts

    @property
    def alert_rules(self):
        """
        Property returning a [`AlertRules`](../alert_rules) for interacting with `/v*/alert-rules` and `/v*/alerts/rules* API endpoints.

        Usage:

            >>> client.alert_rules.v1.add_users(rule_id='test', users=['user-id-1', 'user-id-2'])

        """
        return self._alert_rules

    @property
    def audit_log(self):
        """
        Property returning an [`AuditLogClient`](../audit_log) for interacting with
        `/v*/audit` API endpoints.
        Usage:
            >>> client.audit_log.v1.get_page()
        """
        return self._audit_log

    @property
    def cases(self):
        """
        Property returning a [`CasesClient`](../cases) for interacting with `/v*/cases` API endpoints.

        Usage:

            >>> client.cases.v1.create(name="Test", description="My Description")

        """
        return self._cases

    @property
    def customer(self):
        """
        Property returning a [`CustomerClient`](../customer) for interacting with `/v*/customer` API endpoints.

        Usage:

            >>> client.customer.v1.get()

        """
        return self._customer

    @property
    def departments(self):
        """
        Property returning a [`DepartmentsClient`](../departments) for interacting with `/v*/departments` API endpoints.
        Usage:

            >>> client.departments.v1.get_page()

        """
        return self._departments

    @property
    def devices(self):
        """
        Property returning a [`DevicesClient`](../devices) for interacting with `/v*/devices` API endpoints.
        Usage:

            >>> client.devices.v1.get_page(active=True)

        """
        return self._devices

    @property
    def directory_groups(self):
        """
        Property returning a [`DirectoryGroupsClient`](../directory_groups) for interacting with `/v*/directory-groups` API endpoints.
        Usage:

            >>> client.directory_groups.v1.get_page()

        """
        return self._directory_groups

    @property
    def file_events(self):
        """
        Property returning a [`FileEventsClient`](../file_events) for interacting with `/v*/file-events` API endpoints.
        Usage:

            >>> from incydr import EventQuery
            >>> query = EventQuery(start_date='P30D').equals('file.category', 'Document')

            >>> client.file_events.v2.search(query)

        """
        return self._file_events

    @property
    def files(self):
        """
        Property returning a [`FilesClient`](../files) for interacting with `/v1/files` API endpoints.
        Usage:

            >>> client.files.v1.get_file_by_sha256("sha256 hash", "/path/to/file.extension")
        """
        return self._files

    @property
    def legal_hold(self):
        """
        Property returning a [`LegalHoldClient`](../legal_hold) for interacting with `/v1/legal-hold` API endpoints.

        Usage:

            >>> client.legal_hold.v1.iter_all_matters()
        """
        return self._legal_hold

    @property
    def orgs(self):
        """
        Property returning an ['OrgsClient'](../orgs) for interacting with `/v1/orgs` API endpoints.
        Usage:

            >>> client.orgs.v1.list()
        """
        return self._orgs

    @property
    def sessions(self):
        """
        Property returning a [`SessionsClient`](../sessions) for interacting with `/v*/sessions` API endpoints.
        Usage:

            >>> client.items.v1.get_page(has_alerts=True)

        """
        return self._sessions

    @property
    def trusted_activities(self):
        """
        Property returning a [`TrustedActivitiesClient`](../trusted_activities) for interacting with
        `/v*/trusted-activities` API endpoints.

        Usage:

            >>> client.trusted_activities.v2.get_page()

        """
        return self._trusted_activities

    @property
    def users(self):
        """
        Property returning a [`UsersClient`](../users) for interacting with `/v*/users` API endpoints.
        Usage:

            >>> client.users.v1.get_page(active=True)

        """
        return self._users

    @property
    def risk_profiles(self):
        """
        Property returning a [`RiskProfilesClient`](../risk_profiles) for interacting with
        `/v*/user_risk_profiles` API endpoints.

        Usage:

            >>> client.risk_profiles.v1.get_risk_profile("23")

        """
        return self._risk_profiles

    @property
    def watchlists(self):
        """
        Property returning a [`WatchlistsClient`](../watchlists) for interacting with `/v*/watchlists` API endpoints.
        Usage:

            >>> client.watchlists.v1.get_page()
        """
        return self._watchlists



import click

from _incydr_cli.core import incompatible_with
from _incydr_cli.file_readers import FileOrString

AdvancedQueryAndSavedSearchIncompatible = incompatible_with(
    ["advanced_query", "saved_search"]
)

start_option = click.option(
    "--start",
    default=None,
    help="The beginning of the date range in which to look for file events. Accepts a date/time in yyyy-MM-dd (UTC) or"
    "yyyy-MM-dd HH:MM:SS (UTC+24-hr time) format or a duration in the form of an ISO-duration string (ex. Pass `P7D` to filter for events which occurred in the last week).",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)
end_option = click.option(
    "--end",
    default=None,
    help="The end of the date range in which to look for file events, argument format options are the same as `--start`.",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)
event_action_option = click.option(
    "--event-action",
    default=None,
    help="Filter by the type of file event observed.",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)
username_option = click.option(
    "--username",
    default=None,
    help="Filter by the Code42 username used to sign in to the Code42 app on the device. Null if the file event occurred on a cloud provider.",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)
md5_option = click.option(
    "--md5",
    default=None,
    help="Filter by the MD5 hash of the file contents.",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)
sha256_option = click.option(
    "--sha256",
    default=None,
    help="Filter by the SHA-256 hash of the file contents.",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)
source_category_option = click.option(
    "--source-category",
    default=None,
    help="Filter by the category of where the file originated. For example: Cloud Storage, Email, Social Media.",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)
destination_category_option = click.option(
    "--destination-category",
    default=None,
    help="Filter by the category of the file destination. For example: Cloud Storage, Email, Social Media.",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)
file_name_option = click.option(
    "--file-name",
    default=None,
    help="Filter by the name of the file, including the file extension.",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)
file_directory_option = click.option(
    "--file-directory",
    default=None,
    help="Filter by the file location on the user's device; a forward or backslash must be included at the end of the filepath. Possibly null if the file event occurred on a cloud provider.",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)
file_category_option = click.option(
    "--file-category",
    default=None,
    help="Filter by the categorization of the file that is inferred from MIME type.",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)
risk_indicator_option = click.option(
    "--risk-indicator",
    default=None,
    help="Filter by a list of risk indicators identified for this event. If more than one risk indicator applies to this event, the sum of all indicators determines the total risk score.",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)
risk_severity_option = click.option(
    "--risk-severity",
    default=None,
    help="Filter by the general risk assessment of the event, based on the numeric score.",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)
risk_score_option = click.option(
    "--risk-score",
    default=1,
    help="Filter by risk scores greater than the provided value. The risk score is the sum of the weights for each risk indicator. This score is used to determine the overall risk severity of the event.  "
    "Defaults to 1.  Set to 0 to return all events, including those that have no risk associated with them.",
    cls=AdvancedQueryAndSavedSearchIncompatible,
)

saved_search_option = click.option(
    "--saved-search",
    default=None,
    help="Get events from a saved search with the given ID.  WARNING: Using a saved search is incompatible with other query-building arguments.  Any additional filter options will be ignored.",
    cls=incompatible_with(["advanced_query"]),
)
advanced_query_option = click.option(
    "--advanced-query",
    default=None,
    type=FileOrString(),
    help=(
        "Get events from a raw JSON file events query. Useful for when the provided query parameters do not satisfy your requirements.  "
        "Argument can be passed as a string, read from stdin by passing '-', or from a filename if prefixed with '@',"
        "e.g. '--advanced-query @query.json'. WARNING: Using advanced queries is incompatible with other query-"
        "building arguments.  Any additional filter options will be ignored."
    ),
    cls=incompatible_with(["saved_search"]),
)


def event_filter_options(f):
    f = start_option(f)
    f = end_option(f)
    f = event_action_option(f)
    f = username_option(f)
    f = md5_option(f)
    f = sha256_option(f)
    f = source_category_option(f)
    f = destination_category_option(f)
    f = file_name_option(f)
    f = file_directory_option(f)
    f = file_category_option(f)
    f = risk_indicator_option(f)
    f = risk_severity_option(f)
    f = risk_score_option(f)
    return f



from __future__ import annotations

from datetime import datetime
from typing import List
from typing import Optional
from typing import Union

from pydantic import Extra
from pydantic import Field

from _incydr_sdk.core.models import Model
from _incydr_sdk.core.models import ResponseModel
from _incydr_sdk.enums.trusted_activities import ActivityType
from _incydr_sdk.enums.trusted_activities import Name
from _incydr_sdk.enums.trusted_activities import PrincipalType


class ProviderObject(Model):
    name: Optional[Union[Name, str]] = Field(
        None,
        description="The name of a provider for a specified activity action.\n  ### Supported providers for trusted "
        "activity type and actions\n\n  - `DOMAIN`\n    - `CLOUD_SHARE`\n      - `BOX`\n      - "
        "`GOOGLE_DRIVE` \n      - `ONE_DRIVE`\n    - `CLOUD_SYNC`\n      - `BOX`\n      - `GOOGLE_DRIVE` "
        "\n      - `ICLOUD` \n      - `ONE_DRIVE` \n    - `EMAIL` \n      - `GMAIL`\n      - "
        "`OFFICE_365`\n  - `ACCOUNT_NAME`\n    - `CLOUD_SYNC`\n      - `DROPBOX`\n      - `ONE_DRIVE`\n",
    )


class ActivityAction(Model):
    type: Optional[Union[ActivityType, str]] = Field(
        None,
        description="The type of an activity action.\n### Supported trusted activity types for each activity action "
        "type\n\n- `CLOUD_SHARE` \n  - `DOMAIN`\n- `CLOUD_SYNC` \n  - `DOMAIN`\n  - `ACCOUNT_NAME`\n- "
        "`EMAIL` \n  - `DOMAIN`\n- `FILE_UPLOAD`\n  - `DOMAIN`\n\nNote: `SLACK` and `URL_PATH` do not "
        "need any activity actions specified.\n",
    )
    providers: Optional[List[ProviderObject]] = Field(
        None,
        description="A list of enabled providers for the specified activity action.",
    )


class ActivityActionGroup(Model):
    name: Optional[Union[Name, str]] = Field(
        Name.DEFAULT,
        description="The name of the activity action group. Currently, only `DEFAULT` activity action group is "
        "supported.",
    )
    activityActions: Optional[List[ActivityAction]] = Field(
        None, description="The list of activity actions for an activity action group."
    )


class TrustedActivity(ResponseModel, validate_assignment=True):
    """A model representing an TrustedActivity.

    **Fields**:

    * **activity_action_groups**: `List[ActivityActionGroup]` The list of activity actions associated with the activity.
    * **activity_id**: `str` The unique identifier of the trusted activity.
    * **description**: `str` A description of the trusted activity.
    * **principal_type**: `PrincipalType | None`
    * **activity_type**: `ActivityType | None` The type of the trusted activity.
    * **update_time**: `datetime | None` The time at which the trust activity was last created or modified.
    * **updated_by_principal_id**: `str | None` The unique identifier of the user who last updated the trust activity.
    * **updated_by_principal_name**: `str | None` The username of the user who last updated the trusted activity.
    * **value**: `str` The value of the trusted activity.
    """

    activity_action_groups: Optional[List[ActivityActionGroup]] = Field(
        None,
        description="The list of activity actions associated with the activity.",
        alias="activityActionGroups",
    )
    activity_id: Optional[str] = Field(
        None,
        description="The unique identifier of the trusted activity.",
        alias="activityId",
        frozen=True,
    )
    is_high_value_source: Optional[bool] = Field(
        None,
        description="If the trusted activity represents a high value source",
        alias="isHighValueSource",
    )
    description: Optional[str] = Field(
        None, description="A description of the trusted activity."
    )
    principal_type: Optional[Union[PrincipalType, str]] = Field(
        None, alias="principalType", frozen=True
    )
    type: Optional[Union[ActivityType, str]] = Field(
        None, description="The type of the trusted activity.", alias="type"
    )
    update_time: Optional[datetime] = Field(
        None,
        description="The time at which the trust activity was last created or modified.",
        alias="updateTime",
        frozen=True,
    )
    updated_by_principal_id: Optional[str] = Field(
        None,
        description="The unique identifier of the user who last updated the trust activity.",
        alias="updatedByPrincipalId",
        frozen=True,
    )
    updated_by_principal_name: Optional[str] = Field(
        None,
        description="The username of the user who last updated the trusted activity.",
        alias="updatedByPrincipalName",
        frozen=True,
    )
    value: Optional[str] = Field(None, description="The value of the trusted activity.")


class UpdateTrustedActivity(Model, extra=Extra.ignore):
    activity_action_groups: Optional[List[ActivityActionGroup]] = Field(
        None,
        description="The list of activity action groups for the trusted activity. \n- `DOMAIN` - If array is empty, "
        "existing activity action groups are maintained. \n- `ACCOUNT_NAME` - Atleast 1 activity action "
        "group is required.\n- `SLACK` - No activity action groups are allowed - array must be empty.\n- "
        "`URL_PATH` - No activity action groups are allowed - array must be empty.\n",
        alias="activityActionGroups",
    )
    description: Optional[str] = Field(
        None, description="The description of the trusted activity."
    )
    is_high_value_source: Optional[bool] = Field(
        None,
        description="Whether the trusted activity represents a high value source",
        alias="isHighValueSource",
    )
    type: Optional[Union[ActivityType, str]] = Field(
        None,
        description="The type of the trusted activity.\n\nNote: The `ACCOUNT_NAME` trusted activity type requires "
        "agent version 1.5.0 or later for Incydr \nProfessional, Enterprise, Gov F2, and Horizon product "
        "plans, and Code42 app version 9.0.0 or later for \nIncydr Basic, Advanced, and Gov F1 product "
        "plans.\n",
        alias="type",
    )
    value: Optional[str] = Field(None, description="The value of the trusted activity.")


class TrustedActivitiesPage(ResponseModel):
    """
    A model representing a page of `TrustedActivity` objects.

    **Fields**:

    * **total_count**: `int` Total count of trusted activities found by the query.
    * **trusted_activities**: `List[TrustedActivity]` The list of trusted activities retrieved from the query.
    """

    total_count: Optional[float] = Field(None, alias="totalCount")
    trusted_activities: Optional[List[TrustedActivity]] = Field(
        None,
        description="A list of TrustedActivityResponses",
        alias="trustedActivities",
    )


class QueryTrustedActivitiesRequest(Model):
    page_num: Optional[int]
    page_size: Optional[int]
    activity_type: Optional[str]
    sort_key: Optional[str]
    sort_direction: Optional[str]


class CreateTrustedActivityRequest(Model):
    type: Optional[str]
    value: Optional[str]
    description: Optional[str]
    activityActionGroups: Optional[List[ActivityActionGroup]]



from __future__ import annotations

from csv import DictReader
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from json import JSONDecodeError

import requests
from boltons.jsonutils import JSONLIterator
from pydantic import BaseModel
from pydantic import ConfigDict
from pydantic import model_validator
from pydantic import PrivateAttr
from pydantic import SecretStr
from pydantic import ValidationError


class Model(BaseModel):
    """
    Subclass of pydantic's `BaseModel` to change the `.dict()` and `.json()` methods to dump fields with `by_alias=True`
    as the default.
    """

    def json(
        self,
        *,
        include=None,
        exclude=None,
        by_alias=True,
        exclude_unset=False,
        exclude_defaults=False,
        exclude_none=False,
        **dumps_kwargs,
    ):
        """
        Generate a JSON representation of the model, optionally specifying which fields to include or exclude.

        See [Pydantic docs](https://docs.pydantic.dev/latest/concepts/serialization/#modelmodel_dump_json) for full details.
        """
        return super().model_dump_json(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            **dumps_kwargs,
        )

    def dict(
        self,
        *,
        include=None,
        exclude=None,
        by_alias=True,
        exclude_unset=False,
        exclude_defaults=False,
        exclude_none=False,
    ):
        """
        Generate a dict representation of the model, optionally specifying which fields to include or exclude.

        See [Pydantic docs](https://docs.pydantic.dev/latest/concepts/serialization/#modelmodel_dump) for full details.
        """
        return super().model_dump(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
        )

    @classmethod
    def parse_json_lines(cls, file):
        """
        Accepts an open file-like object in [JSON Lines format](https://jsonlines.org) and returns a generator of
        models parsed from the JSON line by line.
        """
        num = 1
        try:
            for line in JSONLIterator(file):
                try:
                    yield cls(**line)
                    num += 1
                except ValidationError as v_err:
                    raise ValueError(
                        f"Error parsing object on line {num}: {str(v_err)}"
                    )
        except JSONDecodeError:
            raise ValueError(
                f"Unable to parse line {num}. Expecting JSONLines format: https://jsonlines.org"
            )

    model_config = ConfigDict(
        populate_by_name=True,
        use_enum_values=True,
        json_encoders={datetime: lambda dt: dt.isoformat().replace("+00:00", "Z")},
        extra="allow",
    )


class ResponseModel(Model):
    @classmethod
    def parse_response(cls, response: requests.Response):
        try:
            return cls.model_validate_json(response.text)
        except ValidationError as err:
            err.response = response
            raise


def datetime_now_utc_callback():
    return datetime.now(timezone.utc)


class AuthResponse(ResponseModel):
    token_type: str
    expires_in: int
    access_token: SecretStr
    _init_time: datetime = PrivateAttr(default_factory=datetime_now_utc_callback)

    @property
    def expired(self):
        return (datetime.now(timezone.utc) - self._init_time) > timedelta(
            seconds=self.expires_in
        )


class TokenDetails(Model):
    tokenValue: SecretStr
    expiresAt: datetime

    @property
    def expired(self):
        return datetime.now(timezone.utc) >= self.expiresAt


class RefreshTokenAuthResponse(ResponseModel):
    accessToken: TokenDetails
    refreshToken: TokenDetails


class CSVModel(BaseModel):
    """
    Pydantic model class enables multiple aliases to be assigned to a single field value. If the field is required
    then at least one of the aliases must be supplied or validation will fail.

    Useful when parsing CSV data from multiple sources where the expected column header names might vary.

    For example, if a CSV requires a "user" column, which could either be a username or ID:

        class UserCSV(CSVModel):
            user: str = Field(csv_aliases=["user_id", "userId", "username"])
            department: Optional[str]

    Then a CSV could have any of the columns "username", "user_id", or "userId" and the "user" field will be populated
    with the value of that column.

    If a CSV file has multiple alias columns pointing to the same field (e.g. "username" and "userId"), the field will
    be populated by priority of the order of the `csv_aliases` list definition. So in the example above, a CSV that
    has both "username" and "userId" columns, the `model.user` field will be the "userId" CSV value. But because the
    model also allows extra values, "username" will still be accessible on the model at `model.username`.
    """

    model_config = ConfigDict(extra="allow", validate_by_name=True)

    @model_validator(mode="before")
    @classmethod
    def _alias_validator(cls, values):  # noqa
        for name, field_info in cls.model_fields.items():
            if field_info.json_schema_extra:
                aliases = field_info.json_schema_extra.get("csv_aliases", [])
            else:
                aliases = []
            for alias in aliases:
                if alias in values and values[alias]:
                    values[name] = values[alias]
                    break
            else:  # no break
                if field_info.is_required():
                    raise ValueError(
                        f"'{name}' required. Valid column aliases: {aliases}"
                    )

        return values

    @classmethod
    def parse_csv(cls, file):
        first_line = next(file)
        headers = first_line.strip().split(",")
        try:
            cls(**{key: "value" for key in headers})
        except ValidationError as err:
            msg = err.errors()[0]["msg"]
            raise ValueError(f"CSV header missing column: {msg}")

        reader = DictReader(file, fieldnames=headers, restkey="extra")
        for row in reader:
            try:
                # coerce empty columns from "" to None
                row = {k: v or None for k, v in row.items()}
                yield cls(**row)
            except ValidationError as err:
                msg = err.errors()[0]["msg"]
                raise ValueError(f"Missing data on CSV row {reader.line_num}: {msg}")



from __future__ import annotations

from datetime import datetime
from typing import List
from typing import Optional

from pydantic import Field
from rich.markdown import Markdown

from _incydr_sdk.core.models import Model
from _incydr_sdk.core.models import ResponseModel


class User(ResponseModel):
    """
    A model representing a user.

    **Fields**:

    * **legacy_user_id**: `str` - The user ID to use for older console-based APIs that require a device ID.
    * **user_id**: `str` - The globally unique ID (guid) for this user.
    * **username**: `str` - The name the user uses to log in to Code42.
    * **first_name**: `str` - The first (given) name of the user.
    * **last_name**: `str` - The last (family) name of the user.
    * **legacy_org_id**: `str` - The org ID to use for older console-based APIs that require an org Id. If your endpoint domain starts with "console" instead of "api", use this Id for endpoints that require an orgId.
    * **org_id**: `str` - The ID of the Code42 organization this user belongs to.
    * **org_guid**: `str` - The unique org ID of the Code42 organization this user belongs to. This value should be used for actions conducted through the org APIs.
    * **org_name**: `str` - The name of the Code42 organization this user belongs to.
    * **notes**: `str` - Descriptive information about the user.
    * **active**: `bool` - Whether or not the user is enabled.
    * **blocked**: `bool` - Whether or not logins and restores are disabled for the user.
    * **creation_date**: `str` - The date and time the user was created.
    * **modification_date**: `str` - The date and time the user was last modified.
    """

    legacy_user_id: Optional[str] = Field(
        None,
        alias="legacyUserId",
        description="The user ID to use for older console-based APIs that require a "
        'user Id. If your endpoint domain starts with "console" '
        'instead of "api", use this Id for endpoints that require a '
        "userId.",
    )
    user_id: Optional[str] = Field(
        None,
        alias="userId",
        description="A globally unique ID for this user.",
    )
    username: Optional[str] = Field(
        None,
        description="The name the user uses to log in to Code42.",
    )
    first_name: Optional[str] = Field(
        None,
        alias="firstName",
        description="The first (given) name of the user.",
    )
    last_name: Optional[str] = Field(
        None,
        alias="lastName",
        description="The last (family) name of the user.",
    )
    legacy_org_id: Optional[str] = Field(
        None,
        alias="legacyOrgId",
        description="The org ID to use for older console-based APIs that require an "
        'org Id. If your endpoint domain starts with "console" instead '
        'of "api", use this Id for endpoints that require an orgId.',
    )
    org_id: Optional[str] = Field(
        None,
        alias="orgId",
        description="The ID of the Code42 organization this user belongs to.",
    )
    org_guid: Optional[str] = Field(
        None,
        alias="orgGuid",
        description="The unique org ID of the Code42 organization this user belongs to. This value should be used for actions conducted through the org APIs.",
    )
    org_name: Optional[str] = Field(
        None,
        alias="orgName",
        description="The name of the Code42 organization this user belongs to.",
    )
    notes: Optional[str] = Field(
        None,
        description="Descriptive information about the user.",
        table=lambda f: f if f is None else Markdown(f),
    )
    active: Optional[bool] = Field(
        None,
        description="Whether or not the user is enabled.",
    )
    blocked: Optional[bool] = Field(
        None,
        description="Whether or not logins and restores are disabled for the user.",
    )
    creation_date: Optional[datetime] = Field(
        None,
        alias="creationDate",
        description="The date and time the user was created.",
    )
    modification_date: Optional[datetime] = Field(
        None,
        alias="modificationDate",
        description="The date and time the user was last modified.",
    )


class UsersPage(ResponseModel):
    """
    A model representing a page of `User` objects.

    **Fields**:

    * **users**: `List[User]` - The list of `n` number of users retrieved from the query, where `n=page_size`.
    * **total_count**: `int` - Total count of users found by query.
    """

    users: Optional[List[User]] = Field(None, description="A list of users")
    total_count: Optional[int] = Field(
        None, alias="totalCount", description="The total number of users"
    )


class QueryUsersRequest(Model):
    active: Optional[bool]
    blocked: Optional[bool]
    username: Optional[str]
    page: Optional[int] = 1
    pageSize: Optional[int] = 100


class UserRole(ResponseModel):
    """
    A model representing a user role.

    **Fields**:

    * **role_id**: `str` - A role ID.
    * **role_name**: `str` - A role name.
    * **modification_date**: `str` - The date and time this role for the user was last modified.
    * **creation_date**: `str` - The date and time this role for the user was created.
    * **permission_ids**: `str` - The permission IDs associated with this role.
    """

    role_id: Optional[str] = Field(None, alias="roleId", description="A role ID.")
    role_name: Optional[str] = Field(None, alias="roleName", description="A role name.")
    creation_date: Optional[datetime] = Field(
        None,
        alias="creationDate",
        description="The date and time this role for the user was created.",
    )
    modification_date: Optional[datetime] = Field(
        None,
        alias="modificationDate",
        description="The date and time this role for the user was last modified.",
    )
    permission_ids: Optional[List[str]] = Field(
        None,
        alias="permissionIds",
        description="The permission IDs associated with this role.",
    )


class UpdateRolesResponse(ResponseModel):
    """
    A model representing the response to updating a user's role.

    **Fields**:

    * **processed_replacement_role_ids**: `List[str]` - The role IDs processed.
    * **newly_assigned_roles_ids**: `List[str]` - The role IDs newly assigned to the user.
    * **unassigned_roles_ids**: `List[str]` - The role IDs unassigned from the user.
    * **ignored_roles_ids**: `List[str]` - The role IDs ignored.
    """

    processed_replacement_role_ids: Optional[List[str]] = Field(
        None,
        alias="processedReplacementRoleIds",
        description="The role IDs processed.",
    )
    newly_assigned_roles_ids: Optional[List[str]] = Field(
        None,
        alias="newlyAssignedRolesIds",
        description="The role IDs newly assigned to the user.",
    )
    unassigned_roles_ids: Optional[List[str]] = Field(
        None,
        alias="unassignedRolesIds",
        description="The role IDs unassigned from the user.",
    )
    ignored_roles_ids: Optional[List[str]] = Field(
        None,
        alias="ignoredRolesIds",
        description="The role IDs ignored.",
    )


class Permission(Model):
    """
    A model representing a role permission.

    **Fields**:

    * **permission**: `str` - Permission ID.
    * **description**: `str` - Brief description of the permission.
    """

    permission: Optional[str] = Field(None, description="Permission ID.")
    description: Optional[str] = Field(None, description="Permission description.")


class Role(ResponseModel):
    """
    A model representing a role.

    **Fields**:

    * **role_id**: `str` - A role ID.
    * **role_name**: `str` - A role name.
    * **modification_date**: `str` - The date and time this role for the user was last modified.
    * **creation_date**: `str` - The date and time this role for the user was created.
    * **permission_ids**: `str` - The permission IDs associated with this role.
    """

    role_id: Optional[str] = Field(None, alias="roleId", description="A role ID.")
    role_name: Optional[str] = Field(None, alias="roleName", description="A role name.")
    creation_date: Optional[datetime] = Field(
        None,
        alias="creationDate",
        description="The date and time this role for the user was created.",
    )
    modification_date: Optional[datetime] = Field(
        None,
        alias="modificationDate",
        description="The date and time this role for the user was last modified.",
    )
    permissions: Optional[List[Permission]] = Field(
        None,
        description="The permission IDs associated with this role.",
    )



from __future__ import annotations

from datetime import datetime
from typing import List
from typing import Optional

from pydantic import Field

from _incydr_sdk.core.models import Model
from _incydr_sdk.core.models import ResponseModel
from _incydr_sdk.enums.alerts import MessagingMethod


class RuleUser(ResponseModel):
    """
    A model representing a user assigned to a rule.

    **Fields**:

    * **user_id_from_authority**: `str` - A unique Code42 user ID.
    * **user_alias_list**: `List[str]` - List of user aliases associated with the user.
    """

    user_id_from_authority: str = Field(
        None,
        description="User ID from authority.",
        examples=["userIdFromAuthority"],
        alias="userIdFromAuthority",
    )
    aliases: List[str] = Field(
        None,
        description="List of user aliases corresponding to the user ID from the authority.",
        examples=[["userAlias1", "userAlias2"]],
    )


class RuleUsersList(ResponseModel):
    """
    A model representing a list of users associated with a rule, as well as the rule's configuration mode to `INCLUDE` or `EXCLUDE` those users.

    **Fields**:

    * **id**: `str` - Unique ID of the rule.
    * **users**: `List[RuleUser]` - A list of users in the rule's username filter.
    * **mode**: `str` - Indicates how to filter on the user list. Specifies whether to `INCLUDE` or `EXCLUDE` the listed users from the rule.
    """

    id: Optional[str] = Field(
        None,
        description="The id of the rule.",
        examples=["669bb117-d07a-4bd9-9f19-f1d85102cf55"],
    )
    users: Optional[List[RuleUser]] = Field(
        None, description="A list of users in the rule's username filter."
    )
    mode: Optional[str] = Field(
        None,
        description="Indicates how to filter on the user list. INCLUDE or EXCLUDE.",
        table=lambda x: "INCLUDE" if x == "0" else "EXCLUDE",
    )


class CloudSharingDetails(Model):
    observe_all: bool = Field(
        None,
        description="Indicates whether we are watching the cloud sharing connector or not.",
        examples=[True],
        alias="observeAll",
    )
    public_link_share: bool = Field(None, alias="publicLinkShare")
    direct_share: bool = Field(None, alias="directShare")


class DownloadVector(Model):
    observe_all: Optional[bool] = Field(None, alias="observeAll")
    salesforce: Optional[bool] = None
    box: Optional[bool] = None
    google_drive: Optional[bool] = Field(None, alias="googleDrive")
    microsoft_one_drive: Optional[bool] = Field(None, alias="microsoftOneDrive")
    criteria_order: int = Field(None, alias="criteriaOrder")


class EmailVector(Model):
    observe_all: Optional[bool] = Field(None, alias="observeAll")
    gmail: Optional[bool] = None
    microsoft365: Optional[bool] = None
    criteria_order: int = Field(None, alias="criteriaOrder")


class FileUploadCategory(Model):
    observe_all: bool = Field(
        None,
        description="Indicates whether we are watching all of the destinations in the category.",
        examples=[True],
        alias="observeAll",
    )
    destinations: Optional[List[str]] = Field(
        None,
        description="A list of specific destinations to watch within the category.",
        examples=[[]],
    )


class AdvancedSettings(Model):
    observe_uncategorized: bool = Field(None, alias="observeUncategorized")


class RemovableMediaVector(Model):
    is_enabled: bool = Field(
        None,
        description="Indicates whether to watch removable media destinations or not.",
        examples=[True],
        alias="isEnabled",
    )
    criteria_order: int = Field(
        None,
        description="Order in which this vector was added to the rule.",
        examples=[3],
        alias="criteriaOrder",
    )


class FileCategoryFilter(Model):
    categories: Optional[List[str]] = Field(
        None,
        description="List of file categories to alert on.",
        examples=[["Archive", "Pdf", "SourceCode"]],
    )
    criteria_order: int = Field(
        None,
        description="Order in which this vector was added to the rule.",
        examples=[3],
        alias="criteriaOrder",
    )


class FileNameFilter(Model):
    patterns: Optional[List[str]] = Field(
        None,
        description="List of file name patterns to alert on.",
        examples=[["*.cs", "*.sh"]],
    )
    criteria_order: int = Field(
        None,
        description="Order in which this vector was added to the rule.",
        examples=[3],
        alias="criteriaOrder",
    )


class FileTypeMismatchFilter(Model):
    is_enabled: bool = Field(
        None,
        description="Indicates whether or not to alert on file type mismatches only.",
        examples=[True],
        alias="isEnabled",
    )
    criteria_order: int = Field(
        None,
        description="Order in which this vector was added to the rule.",
        examples=[3],
        alias="criteriaOrder",
    )


class RiskIndicatorFilter(Model):
    categories: Optional[List[str]] = None
    indicators: Optional[List[str]] = None
    criteria_order: int = Field(None, alias="criteriaOrder")


class RiskSeverityFilter(Model):
    low: bool
    moderate: bool
    high: bool
    critical: bool
    criteria_order: int = Field(None, alias="criteriaOrder")


class Watchlist(Model):
    id: Optional[str] = None


class NotificationContact(Model):
    is_enabled: bool = Field(
        None,
        description="Indicates whether the notifications for this contact are enabled.",
        examples=[True],
        alias="isEnabled",
    )
    type: Optional[str] = Field(
        None, description="Type of notification.", examples=["EMAIL"]
    )
    address: Optional[str] = Field(
        None,
        description="Address notifications are configured to send to.",
        examples=["myUsername@company.com"],
    )


class CloudSharingVector(Model):
    observe_all: Optional[bool] = Field(
        None,
        description="Indicates whether to watch all cloud sharing connectors.",
        examples=[False],
        alias="observeAll",
    )
    box: Optional[CloudSharingDetails] = Field(
        None,
        description="Configuration for specific cloud sharing monitoring using the Box connector.",
    )
    google_drive: Optional[CloudSharingDetails] = Field(
        None,
        description="Configuration for specific cloud sharing monitoring using the Google Drive connector.",
        alias="googleDrive",
    )
    one_drive: Optional[CloudSharingDetails] = Field(
        None,
        description="Configuration for specific cloud sharing monitoring using the Microsoft OneDrive connector.",
        alias="oneDrive",
    )
    criteria_order: int = Field(
        None,
        description="Order in which this vector was added to the rule.",
        examples=[3],
        alias="criteriaOrder",
    )


class FileUploadVector(Model):
    cloud_storage: Optional[FileUploadCategory] = Field(
        None,
        description="Configuration for which cloud storage destinations to monitor.",
        alias="cloudStorage",
    )
    email: Optional[FileUploadCategory] = Field(
        None, description="Configuration for which email destinations to monitor."
    )
    file_conversion_tool: Optional[FileUploadCategory] = Field(
        None,
        description="Configuration for which file conversion tool destinations to monitor.",
        alias="fileConversionTool",
    )
    messaging: Optional[FileUploadCategory] = Field(
        None, description="Configuration for which messaging destinations to monitor."
    )
    pdf_manager: Optional[FileUploadCategory] = Field(
        None,
        description="Configuration for which PDF manager destinations to monitor.",
        alias="pdfManager",
    )
    productivity_tool: Optional[FileUploadCategory] = Field(
        None,
        description="Configuration for which productivity tool destinations to monitor.",
        alias="productivityTool",
    )
    social_media: Optional[FileUploadCategory] = Field(
        None,
        description="Configuration for which social media destinations to monitor.",
        alias="socialMedia",
    )
    source_code_repository: Optional[FileUploadCategory] = Field(
        None,
        description="Configuration for which source code repository destinations to monitor.",
        alias="sourceCodeRepository",
    )
    web_hosting: Optional[FileUploadCategory] = Field(
        None,
        description="Configuration for which web hosting destinations to monitor.",
        alias="webHosting",
    )
    advanced_settings: Optional[AdvancedSettings] = Field(
        None,
        description="Advanced settings around alerting off of unknown destinations.",
        alias="advancedSettings",
    )
    criteria_order: int = Field(
        None,
        description="Order in which this vector was added to the rule.",
        examples=[3],
        alias="criteriaOrder",
    )


class FileVolumeFilter(Model):
    count_greater_than: int = Field(
        None,
        description="File count threshold that must be exceeded to trigger an alert.",
        examples=[125],
        alias="countGreaterThan",
    )
    operator: Optional[str] = Field(
        None,
        description="Operator to use to combine size and count threshold. AND or OR.",
    )
    size_greater_than_in_bytes: int = Field(
        None,
        description="File size threshold that must be exceeded to trigger an alert.",
        examples=[1024],
        alias="sizeGreaterThanInBytes",
    )
    criteria_order: int = Field(
        None,
        description="Order in which this vector was added to the rule.",
        examples=[3],
        alias="criteriaOrder",
    )


class UsernameFilter(Model):
    mode: Optional[str] = Field(
        None, description="Indicates how to filter on the user list. INCLUDE or EXCLUDE"
    )
    usernames: Optional[List[str]] = Field(
        None,
        description="List of usernames.  Will either alert only on these users or on any user not in list.",
        examples=[["myUsername@company.com", "anotherUsername@company.com"]],
    )
    criteria_order: int = Field(
        None,
        description="Order in which this vector was added to the rule.",
        examples=[3],
        alias="criteriaOrder",
    )


class WatchlistFilter(Model):
    watchlists: Optional[List[Watchlist]] = None
    criteria_order: int = Field(None, alias="criteriaOrder")


class NotificationSettings(Model):
    is_enabled: bool = Field(
        None,
        description="Indicates whether notifications are enabled.",
        examples=[True],
        alias="isEnabled",
    )
    contacts: Optional[List[NotificationContact]] = Field(
        None, description="List of notifications configured."
    )


class RuleVectors(Model):
    cloud_sharing: Optional[CloudSharingVector] = Field(
        None,
        description="Configuration for cloud sharing vectors to monitor.",
        alias="cloudSharing",
    )
    download: Optional[DownloadVector] = None
    email: Optional[EmailVector] = None
    file_upload: Optional[FileUploadVector] = Field(
        None,
        description="Configuration for what file upload vectors to monitor.",
        alias="fileUpload",
    )
    removable_media: Optional[RemovableMediaVector] = Field(
        None,
        description="Configuration for what removable media vectors to monitor.",
        alias="removableMedia",
    )


class RuleFilters(Model):
    file_category: Optional[FileCategoryFilter] = Field(
        None,
        description="Configuration for what file categories to monitor.",
        alias="fileCategory",
    )
    file_name: Optional[FileNameFilter] = Field(
        None,
        description="Configuration for what file names to monitor.",
        alias="fileName",
    )
    file_type_mismatch: Optional[FileTypeMismatchFilter] = Field(
        None,
        description="Configuration for whether to only alert on file mismatches.",
        alias="fileTypeMismatch",
    )
    file_volume: Optional[FileVolumeFilter] = Field(
        None,
        description="Activity thresholds to exceed for an alert to be generated.",
        alias="fileVolume",
    )
    risk_indicator: Optional[RiskIndicatorFilter] = Field(None, alias="riskIndicator")
    risk_severity: Optional[RiskSeverityFilter] = Field(None, alias="riskSeverity")
    username: Optional[UsernameFilter] = Field(
        None, description="Configuration for which users to monitor."
    )
    watchlist: Optional[WatchlistFilter] = None


class EducationSettings(Model):
    lesson_id: Optional[str] = Field(
        None, description="Instructor lesson ID.", alias="lessonId"
    )
    messaging_method: Optional[MessagingMethod] = Field(
        None, description="Messaging method.", alias="messagingMethod"
    )


class RuleDetails(ResponseModel):
    """
    A model representing the details of an alert rule.

    **Fields**:

    * **name**: `str` - Unique name of the rule.
    * **description**: `str` - Description of the rule.
    * **severity**: `str` - [Deprecated field] Indicates severity of rule.
    * **is_enabled**: `bool` - Indicates if the rule is currently enabled.
    * **source**: `str` - [Deprecated field] Indicates source of rule creation.
    * **notifications**: `NotificationSettings` - Notification configuration settings for this rule.
    * **education**: `EducationSettings` - Instructor settings for a rule.
    * **vectors**: `RuleVectors` - The exfiltration vectors to be watched.
    * **filters**: `RuleFilters` - The filters to limit the scope of activity to alert on.
    * **id**: `id` - Unique ID of the rule.
    * **created_at**: `datetime` - Timestamp of when the rule was created.
    * **created_by**: `str` - Individual or service who who created the rule.
    * **modified_at**: `datetime` - Timestamp of when the rule was last modified.
    * **modified_by**: `str` - Individual or service who who last modified the rule.
    * **is_system_rule**: `bool` - Boolean indicator of whether or not the rule is connected to a lens.
    """

    name: Optional[str] = Field(
        None,
        description="Unique name of the rule.",
        examples=["Suspicious File Mismatch Rule"],
    )
    description: Optional[str] = Field(
        None,
        description="Description of the rule.",
        examples=[
            "A rule created to trigger alerts on suspicious file mismatch exfiltration"
        ],
    )
    severity: Optional[str] = Field(
        None,
        description="[Deprecated field] Indicates severity of rule.",
        examples=[""],
    )
    is_enabled: Optional[bool] = Field(
        None,
        description="Indicates whether the rule is enabled or not.",
        examples=[True],
        alias="isEnabled",
    )
    source: Optional[str] = Field(
        None,
        description="[Deprecated field] Indicates source of rule creation.",
        examples=[""],
    )
    notifications: Optional[NotificationSettings] = Field(
        None, description="Notifications configuration settings for this rule."
    )
    education: Optional[EducationSettings] = Field(
        None, description="Education settings for a rule."
    )
    vectors: Optional[RuleVectors] = Field(
        None, description="The exfiltration vectors to be watched."
    )
    filters: Optional[RuleFilters] = Field(
        None, description="The filters to limit the scope of activity to alert on."
    )
    id: Optional[str] = Field(
        None,
        description="Unique identifier of the rule.",
        examples=["ecec5037-aedc-4cf9-aad3-57dcafe1f204"],
    )
    created_at: Optional[datetime] = Field(
        None,
        description="Time when the rule was created.",
        examples=["2022-10-04T17:53:26.4534999Z"],
        alias="createdAt",
    )
    created_by: Optional[str] = Field(
        None,
        description="Individual or service who created the rule.",
        examples=["my-username@company.com"],
        alias="createdBy",
    )
    modified_at: Optional[datetime] = Field(
        None,
        description="Time when the rule was last modified.",
        examples=["2022-10-04T17:53:26.453532Z"],
        alias="modifiedAt",
    )
    modified_by: Optional[str] = Field(
        None,
        description="Individual or service who last modified the rule.",
        examples=["my-username@company.com"],
        alias="modifiedBy",
    )
    is_system_rule: Optional[bool] = Field(
        None,
        description="Boolean indicator of whether or not the rule is connected to a lens.",
        examples=[True],
        alias="isSystemRule",
    )



from typing import List
from typing import Optional

from pydantic import BaseModel
from pydantic import Field

from _incydr_sdk.core.models import ResponseModel


class QueryActorsRequest(BaseModel):
    nameStartsWith: Optional[str] = None
    nameEndsWith: Optional[str] = None
    active: Optional[bool] = None
    pageSize: Optional[int] = None
    page: Optional[int] = None


class Actor(ResponseModel):
    """
    A model representing an Actor.

    **Fields**:

    * **active**: `bool` Whether or not the actor is active.
    * **actor_id**: `str` The unique ID of the actor.
    * **alternate_names**: `List[str]` A list of other names the actor may have.
    * **country**: `str` The actor's country.
    * **department**: `str` The actor's department.
    * **division**: `str` The actor's division.
    * **employee_type**: `str` The actor's employment type, such as if they're a contractor.
    * **end_date**: `str` The actor's end date.
    * **first_name**: `str` The first (given) name of the actor.
    * **in_scope**: `str` The actor's scope state. An actor is considered "in scope" if their activity is monitored in at least one data source.
    * **last_name**: `str` The last (family) name of the actor.
    * **locality**: `str` - The actor's locality (city).
    * **manager_actor_id**: `str` The actor ID of the actor's manager.
    * **name**: `str` The actor's (eg. full username/email) name.
    * **notes**: `str` Notes about the actor.
    * **parentActorId**: `str` The actor ID of this actor's parent actor. (if the actor has a parent).
    * **region**: `str` The actor's region (state).
    * **start_date**: `str` The actor's start date.
    * **title**: `str` The actor's job title.
    """

    active: Optional[bool] = Field(
        None, description="Whether or not the actor is active.", examples=[True]
    )
    actor_id: Optional[str] = Field(
        None, alias="actorId", description="The actor ID.", examples=["112233445566"]
    )
    alternate_names: Optional[List[str]] = Field(
        None,
        alias="alternateNames",
        description="A list of other names the actor may have.",
        examples=[["johnsmith@gmail.com", "john-smith@company.com"]],
    )
    country: Optional[str] = Field(
        None, description="The actor's country", examples=["United States"]
    )
    department: Optional[str] = Field(
        None, description="The actor's department", examples=["Product"]
    )
    division: Optional[str] = Field(
        None, description="The actor's division", examples=["Engineering"]
    )
    employee_type: Optional[str] = Field(
        None,
        alias="employeeType",
        description="The actor's employment, such as if they're a contractor.",
        examples=["full-time"],
    )
    end_date: Optional[str] = Field(
        None,
        alias="endDate",
        description="The actor's end date.",
        examples=["2024-09-18"],
    )
    first_name: Optional[str] = Field(
        None,
        alias="firstName",
        description="The first name of the actor.",
        examples=["Smith"],
    )
    in_scope: Optional[bool] = Field(
        None,
        alias="inScope",
        description="The actor's scope state. An actor is considered 'in scope' if their activity is monitored in at least one data source.",
        examples=[True],
    )
    last_name: Optional[str] = Field(
        None,
        alias="lastName",
        description="The last name of the actor.",
        examples=["John"],
    )
    locality: Optional[str] = Field(
        None, description="The actor's locality (city).", examples=["Minneapolis"]
    )
    manager_actor_id: Optional[str] = Field(
        None,
        alias="managerActorId",
        description="The actor ID of the actor's manager",
        examples=["9988776655"],
    )
    name: Optional[str] = Field(
        None,
        description="The actor's (eg. full username/email) name.",
        examples=["john.smith@gmail.com"],
    )
    notes: Optional[str] = Field(
        None,
        alias="notes",
        description="Notes about the actor.",
        examples=["A super cool person."],
    )
    parent_actor_id: Optional[str] = Field(
        None,
        alias="parentActorId",
        description="The actor ID of this actor's parent actor. (if the actor has a parent).",
        examples=["442244224422"],
    )
    region: Optional[str] = Field(
        None, description="The actor's region.", examples=["Minnesota"]
    )
    start_date: Optional[str] = Field(
        None,
        alias="startDate",
        description="The actor's start date.",
        examples=["2024-09-18"],
    )
    title: Optional[str] = Field(
        None, description="The actor's job title", examples=["Software Engineer"]
    )


class ActorFamily(ResponseModel):
    """
    A model representing an actor family.

    An actor family consists of one or more child actors adopted by a single parent. If an actor is not associated with
    any other actors, the family consists of only the single actor.

    **Fields**:

    * **children**: `List[[Actor][actor-model]]` The list of child actors associated with the parent.
    * **parent**: `[Actor][actor-model]` The parent actor of the family.
    """

    children: List[Actor] = Field(description="The child actors in the family.")
    parent: Actor = Field(description="The parent actor of the family.")


class ActorsPage(ResponseModel):
    """
    A model representing a page of Actors

    **Fields**:

    * **actors**: `List[Actor]` The list of actors returned from the query.
    """

    actors: List[Actor]



# CLI - specific utils.py file to avoid circular imports
from functools import wraps
from signal import getsignal
from signal import SIGINT
from signal import signal

import click
from click import echo
from click import style

from _incydr_sdk.actors.client import ActorNotFoundError


def deprecation_warning(text):
    echo(style(text, fg="red"), err=True)


def user_lookup(client, value):
    """
    Returns the user ID for a given username, or returns the value unchanged if not a username.

    Used with the `user_lookup_callback` method on user args.
    """
    if "@" in str(value):
        # assume username/email was passed
        users = client.users.v1.get_page(username=value).users
        if len(users) < 1:
            raise ValueError(f"User with username '{value}' not found.")
        return users[0].user_id
        # else return ID
    return value


def actor_lookup(client, value):
    """
    Returns the actor ID for a given actor name, or returns the value unchanged if not a username.

    Used with the `actor_lookup_callback` method on user args.
    """
    if "@" in str(value):
        # assume username/email was passed
        try:
            return client.actors.v1.get_actor_by_name(value).actor_id
        except ActorNotFoundError:
            raise ValueError(f"User with username '{value}' not found.")
        # else return ID
    return value


class warn_interrupt:
    """A context decorator class used to wrap functions where a keyboard interrupt could potentially
    leave things in a bad state. Warns the user with provided message and exits when wrapped
    function is complete. Requires user to ctrl-c a second time to force exit.

    Usage:

    @warn_interrupt(warning="example message")
    def my_important_func():
        pass
    """

    def __init__(self, warning="Cancelling operation cleanly, one moment... "):
        self.warning = warning
        self.old_int_handler = None
        self.interrupted = False
        self.exit_instructions = style("Hit CTRL-C again to force quit.", fg="red")

    def __enter__(self):
        self.old_int_handler = getsignal(SIGINT)
        signal(SIGINT, self._handle_interrupts)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.interrupted:
            exit(1)
        signal(SIGINT, self.old_int_handler)

        return False

    def _handle_interrupts(self, sig, frame):
        if not self.interrupted:
            self.interrupted = True
            click.echo(f"\n{self.warning}\n{self.exit_instructions}", err=True)
        else:
            exit()

    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return inner



from typing import Optional

import click

from _incydr_cli import console
from _incydr_cli import logging_options
from _incydr_cli import render
from _incydr_cli.cmds.options.output_options import table_format_option
from _incydr_cli.cmds.options.output_options import TableFormat
from _incydr_cli.core import IncydrCommand
from _incydr_cli.core import IncydrGroup
from _incydr_sdk.core.client import Client
from incydr.models import DirectoryGroup


@click.group(cls=IncydrGroup)
@logging_options
def directory_groups():
    """View directory groups."""


# Did not provide columns option because there are only two columns [groupId, name]
@directory_groups.command("list", cls=IncydrCommand)
@table_format_option
@click.option(
    "--name", default=None, help="Filter by directory groups with a matching name."
)
@logging_options
def list_(
    format_: TableFormat,
    name: Optional[str],
):
    """
    Retrieve directory group information that has been pushed to Code42 from SCIM or User Directory Sync.

    The results can then be used with the watchlists commands to automatically assign users to watchlists by directory group.
    """
    client = Client()
    groups = client.directory_groups.v1.iter_all(name=name)

    if format_ == TableFormat.table:
        render.table(DirectoryGroup, groups)
    elif format_ == TableFormat.csv:
        render.csv(DirectoryGroup, groups)
    elif format_ == TableFormat.json_pretty:
        for group in groups:
            console.print_json(group.json())
    else:
        for group in groups:
            click.echo(group.json())



from typing import List

from pydantic import parse_obj_as
from requests import HTTPError
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from ..exceptions import IncydrException
from .models.response import FileEventsPage
from .models.response import SavedSearch
from _incydr_sdk.queries.file_events import EventQuery


class InvalidQueryException(IncydrException):
    """Raised when the file events search endpoint returns a 400."""

    def __init__(self, query=None, exception=None):
        self.query = query
        self.message = (
            "400 Response Error: Invalid query. Please double check your query filters are valid. "
            "\nTip: Make sure you're specifying your filter fields in dot notation. "
            "\nFor example, filter by 'file.archiveId' to filter by the archiveId field within the file object.)"
        )
        if "problems" in exception.response.json().keys():
            self.message += f"\nRaw problem data from the response: {exception.response.json()['problems']}"
        self.original_exception = exception
        super().__init__(self.message)


class FileEventsV2:
    """
    Client for `/v2/file-events` endpoints.

    Usage example:

        >>> import incydr
        >>> from incydr import EventQuery
        >>>
        >>> client = incydr.Client(**kwargs)
        >>> query = EventQuery(start_date='P30D').equals('file.category', ['Document', 'SourceCode'])
        >>> client.file_events.v2.search(query)
    """

    def __init__(self, parent):
        self._parent = parent
        self._retry_adapter_mounted = False

    def search(self, query: EventQuery) -> FileEventsPage:
        """
        Search for file events.

        If the search response contains a `next_page_token` value, it will automatically be set on the query object's
        `.page_token` field. So if a query results in more total events than your set page size, you can just make
        another call to `.search()` with the same query object to fetch the next page. To get all events from a query,
        continue calling search with the same `EventQuery` object until the response has an empty `.file_events` field.

        See [File Event Pagination][pagination] for more details.

        **Parameters**:

        * **query**: `EventQuery` (required) - The query object to filter file events by different fields.

        **Returns**: A [`FileEventsPage`][fileeventspage-model] object.
        """
        self._mount_retry_adapter()

        try:
            response = self._parent.session.post("/v2/file-events", json=query.dict())
        except HTTPError as err:
            if err.response.status_code == 400:
                raise InvalidQueryException(query=query, exception=err)
            raise err
        page = FileEventsPage.parse_response(response)
        query.page_token = page.next_pg_token
        return page

    def list_saved_searches(self) -> List[SavedSearch]:
        """
        Get all saved searches.

        **Returns**: A list of [`SavedSearch`][savedsearch-model] objects.
        """
        response = self._parent.session.get("/v2/file-events/saved-searches")
        return parse_obj_as(List[SavedSearch], response.json()["searches"])

    def get_saved_search(self, search_id: str) -> SavedSearch:
        """
        Get a single saved search.

        **Parameters**:

        * **search_id**: `str` - The unique ID of the saved search.

        **Returns**: A [`SavedSearch`][savedsearch-model] object.
        """
        response = self._parent.session.get(
            f"/v2/file-events/saved-searches/{search_id}"
        )

        # the api response contains a page with a single search. Returns that single SavedSearch object.
        # the api will return a 404 if a no saved searches matching the id are found.
        page = parse_obj_as(List[SavedSearch], response.json()["searches"])
        return page[0]

    def _mount_retry_adapter(self):
        """Sets custom Retry strategy for FFS url requests to gracefully handle being rate-limited on FFS queries."""
        if not self._retry_adapter_mounted:
            retry_strategy = FFSQueryRetryStrategy(
                status=3,  # retry up to 3 times
                backoff_factor=5,  # if `retry-after` header isn't present, use 5 second exponential backoff
                allowed_methods=[
                    "POST"
                ],  # POST isn't a default allowed method due to it usually modifying resources
                status_forcelist=[
                    429
                ],  # this only handles 429 errors, it won't retry on 5xx
            )
            file_event_adapter = HTTPAdapter(
                pool_connections=200,
                pool_maxsize=4,
                pool_block=True,
                max_retries=retry_strategy,
            )
            self._parent.session.mount(self._parent.settings.url, file_event_adapter)
            self._retry_adapter_mounted = True


class FileEventsClient:
    def __init__(self, parent):
        self._parent = parent
        self._v2 = None

    @property
    def v2(self):
        if self._v2 is None:
            self._v2 = FileEventsV2(self._parent)
        return self._v2


class FFSQueryRetryStrategy(Retry):
    """The forensic search service helpfully responds with a 'retry-after' header, telling us how long until the rate
    limiter is reset. We subclass :class:`urllib3.Retry` just to add a bit of logging so the user can tell why the
    request might look like it's hanging.
    """

    # TODO: Handle debug logging

    def get_retry_after(self, response):
        retry_after = super().get_retry_after(response)
        # if retry_after is not None:
        #     debug.logger.info(
        #         f"Forensic search rate limit hit, retrying after: {int(retry_after)} seconds."
        #     )
        return retry_after

    def get_backoff_time(self):
        backoff_time = super().get_backoff_time()
        # debug.logger.info(
        #     f"Forensic search rate limit hit, retrying after: {backoff_time} seconds."
        # )
        return backoff_time



import os
from os import path

PACKAGE_NAME = "incydr"


def get_user_project_path(*subdirs):
    """
    Converts all args to a consecutive list of subdirectories within the project path.
    If the path does not exist, creates the necessary directories.

    Returns the path on your user dir to /.incydr/[subdir].
    """
    user_project_path = path.join(path.expanduser("~"), f".{PACKAGE_NAME}")
    result_path = path.join(user_project_path, *subdirs)
    if not path.exists(result_path):
        os.makedirs(result_path)
    return result_path



from _incydr_sdk.enums.sessions import ContentInspectionStatuses
from _incydr_sdk.enums.sessions import SessionSeverities
from _incydr_sdk.enums.sessions import SessionStates
from _incydr_sdk.enums.sessions import SortKeys

__all__ = [
    "ContentInspectionStatuses",
    "SessionSeverities",
    "SessionStates",
    "SortKeys",
]
